/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AlienCommand.h"
#include "CppUtil.h"

namespace Alien
{
    //___________________________________________________________________
    const Command::NameMap& Command::_names()
    {
        static const auto names = Base::makeT<NameMap>(
        {
            { Type::None,  "none" },
            { Type::Connect,  QObject::tr( "connect to the grid") },
            { Type::CheckProxy,  QObject::tr( "check GSI proxy validity") },
            { Type::FlushErrors,  QObject::tr( "check failed commands") },
            { Type::Copy,  QObject::tr( "copy selected file") },
            { Type::Open,  QObject::tr( "open selected file") },
            { Type::Remove,  QObject::tr( "remove selected file") },
            { Type::RemoveDirectory,  QObject::tr( "remove selected directory") },
            { Type::DefaultUser, QObject::tr( "query default user name") },
            { Type::DefaultMasterList, QObject::tr( "query master job list for default user") },
            { Type::MasterList, QObject::tr( "query master job list for all users") },
            { Type::ChildrenList, QObject::tr( "query job children list") },
            { Type::Kill, QObject::tr( "kill selected job(s)") },
            { Type::KillMaster, QObject::tr( "kill selected master job(s)") },
            { Type::Resubmit, QObject::tr( "resubmit selected job(s)") },
            { Type::ResubmitMaster, QObject::tr( "resubmit selected job(s)") },
            { Type::Trace, QObject::tr( "query job information") },
            { Type::TraceJdl, QObject::tr( "query job JDL") },
            { Type::TraceAll, QObject::tr( "query job information") },
            { Type::JobFiles, QObject::tr( "query job output files") },
            { Type::ClearQueue, QObject::tr( "clear command queue") },
            { Type::UserList, QObject::tr( "query user list" ) }
        });

        return names;
    }

    //___________________________________________________________________
    int& Command::_counter()
    {
        static int counter( 0 );
        return counter;
    }

    //___________________________________________________________________
    Command::Command( Type type, QStringList&& arguments ):
        Counter( "Command" ),
        id_( _counter()++ ),
        type_( type ),
        arguments_( std::move( arguments ) )
    {}

    #if QT_VERSION < 0x040800
    //___________________________________________________________________
    Command::Command( Type type, std::initializer_list<QString>&& arguments ):
    Command( type, Base::makeT<QStringList>( std::move( arguments ) ) )
    {}
    #endif

    //___________________________________________________________________
    QString Command::user() const
    {
        switch( type_ )
        {
            case Type::DefaultMasterList:
            case Type::MasterList:
            case Type::ChildrenList:
            case Type::Kill:
            case Type::KillMaster:
            case Type::Resubmit:
            case Type::ResubmitMaster:
            case Type::Trace:
            case Type::TraceAll:
            case Type::TraceJdl:
            case Type::ClearQueue:
            return arguments_[0];

            default: return QString();
        }

    }

    //___________________________________________________________________
    int Command::parentJobId() const
    {
        switch( type_ )
        {
            case Type::ChildrenList:
            case Type::Resubmit:
            case Type::Trace:
            case Type::TraceAll:
            case Type::TraceJdl:
            case Type::JobFiles:
            return arguments_[1].toInt();
            default: return -1;
        }
    }

    //___________________________________________________________________
    int Command::jobId() const
    {
        switch( type_ )
        {
            case Type::ChildrenList:
            case Type::Kill:
            case Type::KillMaster:
            case Type::Resubmit:
            case Type::ResubmitMaster:
            case Type::Trace:
            case Type::TraceAll:
            case Type::TraceJdl:
            case Type::JobFiles:
            return arguments_[2].toInt();
            default: return -1;
        }
    }

    //___________________________________________________________________
    QString Command::path() const
    {
        switch( type_ )
        {
            case Type::JobFiles: return arguments_[3];
            default: return QString();
        }
    }

    //___________________________________________________________________
    QString Command::errorString() const
    {
        switch( type_ )
        {
            case Type::CheckProxy:
            return QObject::tr( "No valid GSI proxy was found. "
                "All connections to the grid are likely to fail. "
                "Use <xrdgsiproxy init> to get a valid proxy." );

            case Type::Connect:
            return QObject::tr( "Failed to connect to remote file system." );

            case Type::Copy: return QString( QObject::tr( "Failed to copy file '%1'" ) ).arg( arguments_[0] );
            case Type::Open: return QString( QObject::tr( "Failed to open file '%1'" ) ).arg( arguments_[0] );
            case Type::Remove: return QString( QObject::tr( "Failed to delete file '%1'" ) ).arg( arguments_[0] );
            case Type::RemoveDirectory: return QString( QObject::tr( "Failed to delete folder '%1'" ) ).arg( arguments_[0] );
            case Type::ChildrenList: return QString( QObject::tr( "Failed to query children list for job %1" ) ).arg( jobId() );
            case Type::Kill: return QString( QObject::tr( "Failed to kill job %1" ) ).arg( jobId() );
            case Type::KillMaster: return QString( QObject::tr( "Failed to kill job %1" ) ).arg( jobId() );
            case Type::Resubmit: return QString( QObject::tr( "Failed to resubmit job %1" ) ).arg( jobId() );
            case Type::ResubmitMaster: return QString( QObject::tr( "Failed to resubmit failed children for job %1" ) ).arg( jobId() );
            case Type::Trace: return QString( QObject::tr( "Failed to retrieve information for job %1" ) ).arg( jobId() );
            case Type::TraceAll: return QString( QObject::tr( "Failed to query information for job %1" ) ).arg( jobId() );
            case Type::TraceJdl: return QString( QObject::tr( "Failed to query JDL information for job %1" ) ).arg( jobId() );
            case Type::JobFiles: return QString( QObject::tr( "Failed to query output files for job %1 (path: '%2')" ) ).arg( jobId() ).arg( path() );
            default: return QString( QObject::tr( "Failed to %1" ) ).arg( name() );
        }

    }
}
