#ifndef AlienCommand_h
#define AlienCommand_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "CppUtil.h"
#include "Debug.h"
#include "QOrderedSet.h"

#include <QString>
#include <QStringList>
#include <QHash>
#include <QList>

namespace Alien
{
    class Command: private Base::Counter<Command>
    {

        public:

        //* enumeration for command type
        enum class Type
        {

            None,
            Connect,
            CheckProxy,
            FlushErrors,
            DefaultUser,
            DefaultMasterList,
            MasterList,
            ChildrenList,
            Kill,
            KillMaster,
            Resubmit,
            ResubmitMaster,
            Trace,
            TraceJdl,
            TraceAll,
            JobFiles,
            Copy,
            Open,
            Remove,
            RemoveDirectory,
            UserList,
            ClearQueue
        };

        //* list of commands
        using List = QList<Command>;

        //* constructor
        explicit Command( Type = Type::None, QStringList&& = QStringList() );

        #if QT_VERSION < 0x040800
        //* constructor
        explicit Command( Type, std::initializer_list<QString>&& );
        #endif

        //* destructor
        virtual ~Command() = default;

        //*@name accessors
        //@{

        //* command name
        QString name() const
        { return name( type_ ); }

        //* map Type to command name
        static QString name( Type type )
        { return _names()[type]; }

        //* id
        int id() const
        { return id_; }

        //* priority
        int priority() const
        { return priority_; }

        //* type
        Type type() const
        { return type_; }

        // true if arguments are passed
        bool hasArguments() const
        { return !arguments_.isEmpty(); }

        // true if given argument is passed
        bool hasArgument( int i ) const
        { return arguments_.size() >= i+1; }

        //* arguments
        const QStringList& arguments() const
        { return arguments_; }

        //* arguments
        QString argument( int i ) const
        { return arguments_[i]; }

        //* user
        QString user() const;

        //* master job id
        int parentJobId() const;

        //* job id
        int jobId() const;

        //* path
        QString path() const;

        //* return error message associated to this command, if failed
        QString errorString() const;

        //* error details
        QString errorDetails() const
        { return details_; }

        //@}

        //*@name modifiers
        //@{

        //* priority
        void setPriority( int value )
        { priority_ = value; }

        //* add argument
        Command& addArgument( const QString& argument )
        {
            arguments_.append(argument);
            return *this;
        }

        //* clear arguments
        void clearArguments()
        { arguments_.clear(); }

        //* set error details
        void setErrorDetails( const QString& value )
        { details_ = value; }

        //@}

        private:

        //* static id counter
        static int& _counter();

        //* map types to command name
        using NameMap = QHash<Type,QString>;

        //* names
        static const NameMap& _names();

        //* unique id
        int id_ = 0;

        //* priority
        int priority_ = 0;

        //* type
        Type type_ =Type:: None;

        //* details (in case of errors)
        QString details_;

        //* arguments
        QStringList arguments_;

    };

    using ::qHash;

    //* less than operator
    inline bool operator < ( const Command& first, const Command& second )
    {
        if( first.priority() != second.priority() ) return first.priority() < second.priority();
        else return first.id() < second.id();
    }

    //* equal to operator
    inline bool operator == (const Command& first, const Command& second )
    { return first.priority() == second.priority() && first.id() == second.id(); }

    //* used to sort commands based on priority and type
    class SortedCommand: public Command
    {

        public:

        //* universal constructor
        template< typename... Args >
            explicit SortedCommand( Args&&... args ):
            Command( std::forward<Args>(args)... )
        {}

        using Set = QOrderedSet<SortedCommand>;

    };

    inline bool operator < (const SortedCommand& first, const SortedCommand& second)
    {

        if( first.priority() != second.priority() ) return first.priority() < second.priority();
        if( first.type() != second.type() ) return first.type() < second.type();

        // check arguments
        for( int index = 0; index < std::min( first.arguments().size(), second.arguments().size() ); ++index )
        { if( first.argument(index) != second.argument(index) ) return first.argument(index) < second.argument(index); }

        // check arguments size
        return first.arguments().size() < second.arguments().size();

    };

}

#endif
