/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AlienCommandModel.h"
#include "CppUtil.h"
#include "CustomPixmap.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "IconSize.h"
#include "TimeStamp.h"
#include "XmlOptions.h"

#include <QIcon>

namespace Alien
{

    //_______________________________________________________________
    CommandModel::CommandModel( QObject* parent ):
        ListModel( parent ),
        Counter( "Alien::CommandModel" )
    {

        Debug::Throw( "Alien::CommandModel::CommandModel.\n" );
        updateConfiguration();

    }

    //__________________________________________________________________
    QVariant CommandModel::data( const QModelIndex& index, int role ) const
    {

        // check index
        if( !contains( index ) ) return QVariant();

        // retrieve associated command
        const Command& command( get()[index.row()] );

        // return text associated to command and column
        if( role == Qt::DisplayRole ) {

            switch( index.column() )
            {

                case Id:
                return command.id();

                case Name:
                return command.name();

                case Arguments:
                return command.arguments().join( " " );

                case Priority:
                return command.priority();

                default:
                return QVariant();
            }
        }

        // return icon associated to file
        if( role == Qt::DecorationRole && index.column() == Name )
        { return _icon(command.type()); }

        return QVariant();

    }

    //__________________________________________________________________
    QVariant CommandModel::headerData(int section, Qt::Orientation orientation, int role) const
    {

        if(
            orientation == Qt::Horizontal &&
            role == Qt::DisplayRole &&
            section >= 0 &&
            section < nColumns )
        { return columnTitles_[section]; }

        // return empty
        return QVariant();

    }

    //_________________________________________________________
    void CommandModel::updateConfiguration()
    {

        Debug::Throw( "Alien::CommandModel::updateConfiguration.\n" );

        beginResetModel();

        // reinitialize icons
        _icons().clear();
        _installIcons();

        endResetModel();

    }

    //_____________________________________________________________________
    void CommandModel::_installIcons() const
    {

        Debug::Throw( "Alien::CommandModel::_installIcons" );

        if( !_icons().empty() ) return;

        // type icons
        _icons() = Base::makeT<IconCache>( {
            { Command::Type::Connect, IconEngine::get( IconNames::Connect ) },
            { Command::Type::Connect, IconEngine::get( IconNames::Connect ) },
            { Command::Type::CheckProxy, IconEngine::get( IconNames::Connect ) },
            { Command::Type::DefaultMasterList, IconEngine::get( IconNames::Reload ) },
            { Command::Type::MasterList, IconEngine::get( IconNames::Reload ) },
            { Command::Type::ChildrenList, IconEngine::get( IconNames::Reload ) },
            { Command::Type::DefaultUser, IconEngine::get( IconNames::Users ) },
            { Command::Type::UserList, IconEngine::get( IconNames::Users ) },
            { Command::Type::Kill, IconEngine::get( IconNames::Kill ) },
            { Command::Type::KillMaster, IconEngine::get( IconNames::Kill ) },
            { Command::Type::ClearQueue, IconEngine::get( IconNames::Delete ) },
            { Command::Type::Copy, IconEngine::get( IconNames::Paste ) },
            { Command::Type::Remove, IconEngine::get( IconNames::Delete ) },
            { Command::Type::RemoveDirectory, IconEngine::get( IconNames::Delete ) },
            { Command::Type::Trace, IconEngine::get( IconNames::Information ) },
            { Command::Type::TraceJdl, IconEngine::get( IconNames::Information ) },
            { Command::Type::TraceAll, IconEngine::get( IconNames::Information ) }
        });

        return;
    }

    //_______________________________________________
    CommandModel::IconCache& CommandModel::_icons()
    {
        static IconCache cache;
        return cache;
    }

    //____________________________________________________________
    const QIcon& CommandModel::_icon( Command::Type type )
    {

        IconCache::iterator iter = ( _icons().find( type ) );
        if( iter != _icons().end() ) return iter.value();
        else return _icons().insert( type, CustomPixmap( IconSize::get( IconSize::Medium ), CustomPixmap::Flag::Transparent ) ).value();

    }

}
