/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AlienInterface.h"

#include "Application.h"
#include "Command.h"
#include "CppUtil.h"
#include "Debug.h"
#include "File.h"
#include "FileRecordProperties.h"
#include "JobSet.h"
#include "QOrderedSet.h"
#include "Singleton.h"
#include "TextFormat.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QDesktopServices>
#include <QRegExp>
#include <QTextStream>
#include <QUrl>

namespace Alien
{

    //_______________________________________________
    Interface::Interface( QObject *parent ):
        QObject( parent ),
        Counter( "Alien::Interface" )
    {
        Debug::Throw("Alien::Interface::Interface.\n");
        connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
        _updateConfiguration();
    }

    //_______________________________________________
    void Interface::findCertificates()
    {

        Debug::Throw( "Alien::Interface::findCertificates.\n" );

        // check if option is already set and valid
        if( XmlOptions::get().contains( "CERTIFICATES_PATH" ) &&
            File( XmlOptions::get().raw( "CERTIFICATES_PATH" ) ).exists() &&
            File( XmlOptions::get().raw( "CERTIFICATES_PATH" ) ).isAbsolute() )
        { return; }

        File alienPath;

        // try load from gbbox path
        if( XmlOptions::get().contains( "GBBOX" ) && File( XmlOptions::get().raw( "GBBOX" ) ).isAbsolute() )
        {
            File gbboxPath( File( XmlOptions::get().raw( "GBBOX" ) ).path() );
            alienPath = File( "../.." ).addPath( gbboxPath ).canonicalName();
        }

        if( !alienPath.exists() )
        {

            // try load using GSHELL_ROOT
            const QProcessEnvironment environment( QProcessEnvironment::systemEnvironment() );
            if( environment.contains( "GSHELL_ROOT" ) )
            { alienPath = File( ".." ).addPath( File( environment.value( "GSHELL_ROOT" ) ) ).canonicalName(); }

            // try load using ALIEN_ROOT environment
            if( !alienPath.exists() && environment.contains( "ALIEN_ROOT" ) )
            { alienPath = File( environment.value( "ALIEN_ROOT" ) ).canonicalName(); }

        }

        if( !alienPath.exists() ) return;

        // possible path for certificates, relative to alien path
        const auto pathList = Base::makeT<File::List>(
        {
            File( "globus/share/certificates" ),
            File( "api/share/certificates" )
        });

        // get certificates path
        File certificatesPath;
        for( auto path:pathList )
        {
            const File tmp( path.addPath( alienPath ).canonicalName() );
            if( tmp.exists() )
            {
                certificatesPath = tmp;
                break;
            }
        }

        // assign
        if( certificatesPath.exists() )
        { XmlOptions::get().setRaw( "CERTIFICATES_PATH", certificatesPath ); }

        return;

    }

    //__________________________________________________________________________________
    void Interface::addCommand( const Command& command )
    {

        // keep empty state
        const bool empty( commands_.isEmpty() );

        // doe nothing if command is not effectively added (duplicate)
        emit commandAdded( command );
        if( !commands_.append( command ) )
        {
            emit commandCompleted( command );
            return;
        }

        Debug::Throw() << "Alien::Interface::addCommand - " << command.name() << " (" << command.id() << ")" << endl;


        // emit signal

        // run command if there is none pending
        if( empty )
        {
            emit busyStateChanged( true );
            emit busy();
            _runCommand();
        }
    }

    //__________________________________________________________________________________
    void Interface::addCommands( const Command::List& commands )
    {

        Debug::Throw( "Alien::Interface::addCommands.\n" );

        // list of added commands
        Command::List addedCommands;

        const bool empty( commands_.isEmpty() );
        for( const auto& command:commands )
        {
            if( commands_.append( command ) )
            {

                Debug::Throw() << "Alien::Interface::addCommands - " << command.name() << " (" << command.id() << ")" << endl;
                addedCommands << command;

            } else {

                emit commandAdded( command );
                emit commandCompleted( command );

            }

        }

        // check if commands have been effectively added
        if( addedCommands.isEmpty() ) return;

        emit commandsAdded( addedCommands );
        if( empty )
        {
            emit busyStateChanged( true );
            emit busy();
            _runCommand();
        }

    }

    //__________________________________________________________________________________
    void Interface::abortAll()
    {

        Debug::Throw( "Alien::Interface::abortAll.\n" );

        // abort all commands but the current
        for( const auto& command:commands_.commands() )
        { if( commands_.abort( command ) ) emit commandCompleted( command ); }

        // kill process if any. this should also abort current command
        if( process_ && process_->state() != QProcess::NotRunning )
        { process_->kill(); }

    }

    //__________________________________________________________________________________
    void Interface::abortCommands( const Command::List& commands )
    {
        Debug::Throw( "Alien::Interface::abortCommands.\n" );
        for( const auto& command:commands )
        { if( commands_.abort( command ) ) emit commandCompleted( command ); }
    }

    //____________________________________________________________
    void Interface::clearQueue( const QString& user, const Job::IdSet& jobs )
    {

        Debug::Throw() << "Alieninterface::clearQueue - user: " << user << " jobs: " << jobs.size() << endl;

        // list of commands to be aborted.
        Command::List abortedCommands;

        // get all commands from queue
        SortedCommand::Set commands( commands_.commands() );
        for( const auto& command:commands )
        {
            // check user
            if( command.user() != user ) continue;

            // check job and parent id, for removal
            if(
                ( command.jobId() >= 0 && !jobs.contains( command.jobId() ) ) ||
                ( command.parentJobId() >= 0 && !jobs.contains( command.parentJobId() ) ) )
            { abortedCommands.append( command ); }

        }

        if( !abortedCommands.isEmpty() )
        { abortCommands( abortedCommands ); }

    }

    //__________________________________________________________________________________
    bool Interface::_runCommand()
    {

        // do nothing if paused
        if( paused_ ) return false;

        // do nothing if no commands are in the pile
        if( commands_.isEmpty() )
        {

            Debug::Throw( "Alien::Interface::_runCommand - ready \n" );

            QString buffer = tr( "Last update: %1" ).arg( TimeStamp::now().toString( TimeStamp::Format::TimeLong ) );
            emit messageAvailable( buffer );

            QMetaObject::invokeMethod( this, "ready", Qt::QueuedConnection );
            emit busyStateChanged( false );

            return false;

        }

        // front command
        Command command( commands_.front() );
        Debug::Throw() << "Alien::Interface::_runCommand - command: " << command.name() << " (" << command.id() << ")" << endl;

        // reset ignore error flag
        if( command.type() != Command::Type::FlushErrors )
        { ignoreErrors_ = false; }

        /*
        don't run command if it is associated to a user that is not in the
        list of active users any more
        */
        if(
            (
            command.type() == Command::Type::DefaultMasterList ||
            command.type() == Command::Type::MasterList ||
            command.type() == Command::Type::ChildrenList ||
            command.type() == Command::Type::Trace ) &&
            !activeUsers_.contains( command.user() ) )
        {
            emit commandCompleted( command );
            commands_.removeFirst();
            return _runCommand();
        }

        // reset buffer
        _resetBuffers();
        process_.reset();

        // log commands
        const bool logCommand( XmlOptions::get().get<bool>( "SHOW_COMMANDS_IN_LOG" ) );

        // take action depending on command type
        switch( command.type() )
        {

            case Command::Type::CheckProxy:
            {

                emit logAvailable( tr( "Checking GSI proxy\n" ), Format::Bold );
                const auto arguments = Base::makeT<QStringList>( { XmlOptions::get().raw("GSI_PROXY" ), "info" });
                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;

            }

            case Command::Type::Connect:
            {

                emit logAvailable( tr( "Trying to connect to remote filesystem\n" ), Format::Bold );
                const auto arguments = Base::makeT<QStringList>({
                    XmlOptions::get().raw("GBBOX" ),
                    "connect",
                    XmlOptions::get().raw( "ALIEN_HOST" ),
                    QString::number( XmlOptions::get().get<int>( "ALIEN_PORT" ) ),
                    XmlOptions::get().raw( "ALIEN_USER" )
                });

                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;

            }

            case Command::Type::FlushErrors:
            {

                if( ignoreErrors_ ) clearFailedCommands();
                else {
                    if( !failedCommands_.isEmpty() )  emit commandsFailed( failedCommands_ );
                    if( !failedCommands_.isEmpty() )
                    {
                        emit logAvailable( tr( "Trying to recover failed commands\n" ), Format::Bold );
                        auto commands( failedCommands_ );
                        clearFailedCommands();
                        addCommands( commands );
                        addCommand( Command( Command::Type::FlushErrors ) );

                    }

                }

                break;

            }

            case Command::Type::DefaultUser:
            {

                emit logAvailable( tr( "Requesting default user\n" ), Format::Bold );
                emit messageAvailable( tr( "Requesting default user" ) );

                const auto arguments = Base::makeT<QStringList>({
                    XmlOptions::get().raw("GBBOX" ),
                    "whoami"
                });

                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;

            }

            case Command::Type::UserList:
            {

                emit logAvailable( tr( "Requesting user list\n" ), Format::Bold );
                emit messageAvailable( tr( "Requesting user list" ) );

                const auto arguments = Base::makeT<QStringList>( {
                    XmlOptions::get().raw("GBBOX" ),
                    "who"
                });

                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;

            }

            case Command::Type::Copy:
            case Command::Type::Open:
            {

                const QString remote( command.argument(0) );
                const QString local( command.argument(1) );

                // emit message
                QString buffer = tr( "Copying file %1" ).arg( remote );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                // build command
                const auto arguments = Base::makeT<QStringList>( {
                    XmlOptions::get().raw("ALIEN_CP"),
                    QString( "alien:" ) + remote,
                    QString( "file:" ) + local
                });

                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );

                return true;
            }

            case Command::Type::Remove:
            {
                const QString remote( command.argument(0) );

                // emit message
                QString buffer = tr( "Removing file %1" ).arg( remote );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                // build command
                const auto arguments = Base::makeT<QStringList>( {
                    XmlOptions::get().raw("GBBOX"),
                    QString( "rm" ),
                    remote
                });

                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );

                return true;
            }

            case Command::Type::RemoveDirectory:
            {
                const QString remote( command.argument(0) );

                // emit message
                QString buffer = tr( "Removing folder %1" ).arg( remote );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                // build command
                const auto arguments = Base::makeT<QStringList>({
                    XmlOptions::get().raw("GBBOX"),
                    QString( "rmdir" ),
                    remote
                });

                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );

                return true;
            }

            case Command::Type::DefaultMasterList:
            {

                // get user and check
                const QString user( command.user() );
                if( user.isEmpty() ) break;

                // emit message
                QString buffer = tr( "Requesting job list for user %1" ).arg( user );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                // build command
                const auto arguments = Base::makeT<QStringList>({ XmlOptions::get().raw("GBBOX" ), "ps", "-A" });
                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;

            }

            case Command::Type::MasterList:
            {

                // get user and check
                const auto users( command.arguments() );
                if( users.isEmpty() ) break;

                // emit message
                QString buffer = (users.size() == 1) ?
                    tr( "Requesting job list for user %1" ).arg( users.front() ):
                    tr( "Requesting job list for users %1" ).arg( users.join( " " ) );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                // build command
                const auto arguments = Base::makeT<QStringList>( { XmlOptions::get().raw("GBBOX" ), "ps", "-A", "-a" });
                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;

            }

            case Command::Type::ChildrenList:
            {

                // check arguments
                const QString user( command.user() );
                const int id( command.jobId() );

                // emit message
                QString buffer = tr( "Requesting children list for job %1" ).arg( id );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                // create command
                const auto arguments = Base::makeT<QStringList>( {
                    XmlOptions::get().raw("GBBOX"), "top",
                    "-user", user,
                    "-split", QString::number( id ),
                    "-all_status"
                });

                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;

            }

            case Command::Type::Kill:
            {

                // get user and check
                const int id( command.jobId() );

                // emit message
                QString buffer = tr( "Killing job %1" ).arg( id );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                const auto arguments = Base::makeT<QStringList>({ XmlOptions::get().raw( "GBBOX" ), "kill", QString::number( id ) });
                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;
            }

            case Command::Type::KillMaster:
            {

                // get id
                const int id( command.jobId() );
                if( id < 0 ) break;

                // emit message
                QString buffer = tr( "Killing job %1" ).arg( id );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                // build command
                auto arguments = Base::makeT<QStringList>( { XmlOptions::get().raw( "GBBOX" ), "masterJob", QString::number(id) });

                // add status
                if( command.hasArgument(3) )
                { arguments.append( Base::makeT<QStringList>({ "-status", command.argument(3) }) ); }

                arguments.append( "kill" );

                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;

            }

            case Command::Type::Resubmit:
            {

                // get user and check
                const int id( command.jobId() );
                if( id < 0 ) break;

                // emit message
                QString buffer = tr( "Resubmitting job %1" ).arg( id );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                const auto arguments = Base::makeT<QStringList>( { XmlOptions::get().raw( "ALIEN_RESUBMIT" ), QString::number( id ) });
                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;
            }

            case Command::Type::ResubmitMaster:
            {

                // check arguments
                if( !command.hasArguments() ) break;

                // get user and check
                const int id( command.jobId() );
                if( id < 0 ) break;

                // get status from arguments, if any
                const QString status( command.hasArgument(3) ? command.argument(3):"ERROR_ALL" );

                // run and emit signal
                QString buffer = tr( "Resubmitting failed jobs for job %1 (status: %2)" ).arg( id ).arg( status );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                const auto arguments = Base::makeT<QStringList>({ XmlOptions::get().raw( "GBBOX" ), "masterJob", QString::number( id ), "-status", status, "resubmit" });
                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // start process
                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;

            }

            case Command::Type::Trace:
            case Command::Type::TraceAll:
            {

                // get id
                const int parentId( command.parentJobId() );
                const int id( command.jobId() );

                // emit message
                QString buffer = tr( "Requesting info for job %1" ).arg( id );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                // build command
                auto arguments = Base::makeT<QStringList>( {
                    XmlOptions::get().raw("ALIEN_PS"),
                    "-trace",
                    QString::number( id )
                });

                // add output selection string based on command type
                if( command.type() == Command::Type::Trace )
                {

                    // request all state lines
                    arguments.append( "state" );

                    // all so request "submission" lines, to get list of children
                    if( parentId < 0 ) arguments.append( "submit" );

                } else if( command.type() == Command::Type::TraceAll ) arguments.append( "all" );

                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;

            }

            case Command::Type::TraceJdl:
            {

                // get id
                const int id( command.jobId() );

                // emit message
                QString buffer = tr( "Requesting JDL for job %1" ).arg( id );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                // build command
                auto arguments = Base::makeT<QStringList>( {
                    XmlOptions::get().raw("ALIEN_PS"),
                    "-jdl",
                    QString::number( id )
                });

                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;

            }

            case Command::Type::JobFiles:
            {

                // get id
                const int id( command.jobId() );
                const QString path( command.path() );

                // emit message
                QString buffer = tr( "Requesting output files list for job %1" ).arg( id );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );

                // build command
                const auto arguments = Base::makeT<QStringList>({
                    XmlOptions::get().raw( "GBBOX" ),
                    "ls", "-n", "-l", path
                });

                if( logCommand ) { emit logAvailable( arguments.join( " " ) + "\n" ); }

                // run and emit signal
                _newProcess()->start( arguments );
                emit commandStarted( command );
                return true;

            }

            case Command::Type::ClearQueue:
            {

                const QString user( command.user() );
                Q_ASSERT( !user.isEmpty() );

                // emit message
                QString buffer = tr( "Clearing command queue for user %1" ).arg( user );
                emit logAvailable( buffer + "\n", Format::Bold );
                emit messageAvailable( buffer );
                break;
            }

            default:
            break;
        }

        emit commandCompleted( command );
        commands_.removeFirst();
        return _runCommand();

    }

    //______________________________________________________________
    void Interface::_commandCompleted( int returnCode, QProcess::ExitStatus status )
    {

        /*
        if there is no current command in the queue (in case of abort)
        just jump to the next macro
        */
        if( commands_.isEmpty() )
        {
            _runCommand();
            return;
        }

        // fetch command
        Command command( commands_.front() );
        Debug::Throw()
            << "Alien::Interface::_commandCompleted -"
            << " command: " << command.name() << " (" << command.id() << ")"
            << " status: " << status
            << " code: " << returnCode << endl;

        // check connection state
        const bool ignoreDisconnect(
            command.type() == Command::Type::CheckProxy ||
            command.type() == Command::Type::FlushErrors );
        const bool connected( ignoreDisconnect || _checkConnection() );

        // force return code to zero for MasterList
        if(
            (
            command.type() == Command::Type::DefaultMasterList ||
            command.type() == Command::Type::MasterList ||
            command.type() == Command::Type::ChildrenList ||
            command.type() == Command::Type::UserList )
            && returnCode == 1 ) returnCode = 0;

        // check status
        if( status != QProcess::NormalExit || ( process_ && process_->error() == QProcess::FailedToStart ) || returnCode != 0 || !connected )
        {

            const QString buffer = command.errorString();
            emit messageAvailable( buffer );
            emit logAvailable( buffer + "\n\n", Format::Italic|Format::Bold );

            // store failed command
            command.setErrorDetails( stderrBuffer_.text() );
            failedCommands_.append( command );

            emit commandCompleted( command );
            commands_.removeFirst();

            // add FlushError command
            if( !_ignoreError( command, stderrBuffer_.text() ) )
            { addCommand( Command( Command::Type::FlushErrors ) ); }

            _runCommand();
            return;

        }

        // add (cosmetics) empty line to log.
        emit logAvailable( "\n" );

        // run action depending on current type
        switch( command.type() )
        {

            case Command::Type::CheckProxy:
            {
                _processProxyInfo();
                break;
            }

            case Command::Type::DefaultUser:
            {

                _processDefaultUser();
                break;
            }

            case Command::Type::UserList:
            {

                _processUserList();
                break;
            }

            case Command::Type::Open:
            {

                // run external application, if provided
                if( command.hasArgument(2) )
                {

                    ( Base::Command( command.argument(2) ) << command.argument(1) ).run();

                } else {

                    QDesktopServices::openUrl( QUrl::fromEncoded( QString( "file://%1" ).arg( command.argument(1) ).toLatin1() ) );

                }

                break;
            }

            case Command::Type::DefaultMasterList:
            case Command::Type::MasterList:
            {
                _processMasterList( command.arguments() );
                break;
            }

            case Command::Type::ChildrenList:
            {
                Job job;
                job.assign( Job::User, command.user() )
                    .assign( Job::Id, command.jobId() );
                _processChildrenList( job );
                break;
            }

            case Command::Type::Trace:
            case Command::Type::TraceAll:
            {
                Job job;
                job.setParentId( command.parentJobId() )
                    .assign( Job::User, command.user() )
                    .assign( Job::Id, command.jobId() );
                _processJobInformation( job, command.type() == Command::Type::TraceAll );
                break;
            }

            case Command::Type::TraceJdl:
            {
                Job job;
                job.setParentId( command.parentJobId() )
                    .assign( Job::User, command.user() )
                    .assign( Job::Id, command.jobId() );
                _processJdl( job );
                break;
            }

            case Command::Type::JobFiles:
            {
                Job job;
                job.setParentId( command.parentJobId() )
                    .assign( Job::User, command.user() )
                    .assign( Job::Id, command.jobId() );
                _processJobFiles( job, File( command.argument(3) ) );
                 break;
            }

            default: break;

        }

        // emit signal, remove from pile and run next command if any
        emit commandCompleted( command );
        commands_.removeFirst();
        _runCommand();

        return;

    }

    //____________________________________________________________
    void Interface::_error( QProcess::ProcessError error )
    {

        Q_ASSERT( !commands_.isEmpty() );
        const Command command( commands_.front() );
        Debug::Throw() << "Alien::Interface::_error - command: " << command.name() << " (" << command.id() << ")" << endl;

        if( error == QProcess::FailedToStart )
        {

            /*
            in case the process failed to start, _commandComppleted is not called
            one needs to call it manually in order to proceed with command list
            */
            _commandCompleted( 1, QProcess::NormalExit );

        } else {

            const QString buffer = command.errorString();
            emit messageAvailable( buffer );
            emit logAvailable( buffer+"\n\n", Format::Italic|Format::Bold );

        }

        return;

    }

    //________________________________________________________________________
    bool Interface::_checkConnection()
    {

        Debug::Throw() << "Alien::Interface::_checkConnection -"
            << " connected: " << connected_
            << " wasConnected: " << wasConnected_
            << endl;

        // Warning [Mon Oct 29 09:39:48 2012] GCLIENT::Reconnect
        // Warning [Mon Oct 29 09:39:48 2012] GCLIENT::COMMAND command failed
        QRegExp regExp1( "(Error)|(Warning)\\s+\\[.+\\]\\s+GCLIENT::Reconnect" );
        QRegExp regExp2( "(Error)|(Warning)\\s+\\[.+\\]\\s+GCLIENT::COMMAND command failed" );
        if( stderrBuffer_.text().indexOf( regExp1 ) >= 0 && stderrBuffer_.text().indexOf( regExp2 ) >= 0 )
        {

            if( connected_ )
            {
                connected_ = false;
                emit connectionLost( stderrBuffer_.text() );
                emit connectionStateChanged( false );
            }

            return false;

        } else if( !( connected_ && wasConnected_ ) ) {

            emit connectionStateChanged( true );
            wasConnected_ = true;
            connected_ = true;

        }

        return true;

    }

    //________________________________________________________________________
    void Interface::_processProxyInfo()
    {
        Debug::Throw( "Alien::Interface::_processProxyInfo.\n" );

        // get current command
        Command command( commands_.front() );
        Q_ASSERT( command.type() == Command::Type::CheckProxy );

        // proxy validity
        bool valid( false );

        // time left   : 6h:52m:40s
        QRegExp regexp( "time\\s+left\\s+:\\s+(\\d+)h:(\\d+)m:(\\d+)s" );
        if( stderrBuffer_.text().indexOf( regexp ) >= 0 )
        {

            const int totalTime = 3600*regexp.cap(1).toInt() + 60*regexp.cap(2).toInt() + regexp.cap(3).toInt();
            valid = ( totalTime > 0 );

        }

        // if invalid emit signal
        if( !valid )
        {
            command.setErrorDetails( stderrBuffer_.text() );
            failedCommands_.append( command );
            addCommand( Command( Command::Type::FlushErrors ) );
        }

    }

    //________________________________________________________________________
    void Interface::_processDefaultUser()
    {

        Debug::Throw( "Alien::Interface::_processDefaultUser.\n" );
        QString defaultUser;

        // look for default user first in stdout then stderr
        for( auto buffer:{ stdoutBuffer_.text(), stderrBuffer_.text() } )
        {
            QTextStream stream( &buffer, QIODevice::ReadOnly );
            QStringList lines;
            QString line;
            while( !( line = stream.readLine() ).isNull() ) lines.append( line );

            // read default user
            if( !lines.isEmpty() && !_isError( line = lines.back() ) )
            { defaultUser = line.trimmed(); }

            if( !defaultUser.isEmpty() ) break;
        }

        // build path from user name
        if( defaultUser.isEmpty() )
        {

            emit messageAvailable( tr( "Failed to retrieve default user" ) );

        } else {

            XmlOptions::get().setRaw( "ALIEN_USER", defaultUser );
            emit messageAvailable( "Default user available" );
            emit defaultUserAvailable( defaultUser );

        }

    }

    //________________________________________________________________________
    void Interface::_processUserList()
    {

        Debug::Throw( "Alien::Interface::_processUserList.\n" );
        UserSet users;
        users.setNeedsDefault( false );

        // parse output
        QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
        QString line;

        QOrderedSet<QString> orderedUsers;
        while( !( line = stream.readLine() ).isNull() )
        { if( !( line = line.trimmed() ).isEmpty() ) orderedUsers.insert( line ); }

        users.append( orderedUsers.toList() );

        emit messageAvailable( "User list available" );
        emit userListAvailable( users );

    }

    //________________________________________________________________________
    void Interface::_processMasterList( const QStringList& users )
    {

        Debug::Throw( "Alien::Interface::_processMasterList.\n" );

        // list of found jobs
        JobMap jobs;
        JobMap killedJobs;

        // generate one empty list per user
        for( const auto& user:users )
        { jobs.insert( user, JobSet() ); }

        // parse output
        QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
        QString line;
        while( !( line = stream.readLine() ).isNull() )
        {

            if( line.isEmpty() ) continue;

            // create job from line
            Job job(line.trimmed(), Job::ParseMaster);

            // check validity
            if( !job.isValid() ) continue;

            // check user
            const QString user( job.get( Job::User ) );
            if( !users.contains( user ) ) continue;

            // store job in relevant job sets
            if( job.isKilled() )
            {

                killedJobs[user].insert( job );
                if( !hideKilledJobs_ ) jobs[user].insert( job );

            } else {

                // append to job sest
                jobs[user].insert( job );

            }

        }

        // store killed jobs
        for( JobMap::const_iterator iter = killedJobs.constBegin(); iter != killedJobs.constEnd(); ++iter )
        { killedJobs_[iter.key()] = iter.value(); }

        emit messageAvailable( "Master job list available" );

        for( JobMap::const_iterator iter = jobs.constBegin(); iter != jobs.constEnd(); ++iter )
        { emit masterListAvailable( iter.key(), iter.value() ); }

    }

    //________________________________________________________________________
    void Interface::_processChildrenList( const Job& master )
    {

        Debug::Throw( "Alien::Interface::_processJobList.\n" );

        // list of found jobs
        JobSet jobs;
        JobSet killedJobs;

        // get user
        const QString user( master.get( Job::User ) );

        // parse output
        QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
        QString line;
        while( !( line = stream.readLine() ).isNull() )
        {

            if( line.isEmpty() ) continue;

            // create job from line
            Job job(line.trimmed(), Job::ParseChildren);

            // assign user and parent
            job.assign( Job::User, user );
            job.setParentId( master.id() );

            // check validity
            if( !job.isValid() )
            {
                Debug::Throw() << "Alien::Interface::_processChildrenList - invalid job: " << line << endl;
                continue;
            }

            Debug::Throw() << "Alien::Interface::_processChildrenList - user: " << job.get( Job::User ) << " parent:" << job.parentId() << " id: " << job.id() << endl;

            // store job in relevant job sets
            if( job.isKilled() )
            {

                killedJobs.insert( job );
                if( !hideKilledJobs_ ) jobs.insert( job );

            } else  jobs.insert( job );

        }

        // store killed jobs
        killedJobs_[user] = killedJobs;

        emit childrenListAvailable( master, jobs );

    }

    //________________________________________________________________________
    void Interface::_processJobFiles( const Job& job, const File& path )
    {
        Debug::Throw( "Alien::Interface::_processJobFiles" );

        // list of files
        FileRecord::List files;

        // regular expression
        QRegExp fileRegExp(
            "(\\S+)\\s+"                                      // permissions
            "(\\w+)\\s+(\\w+)\\s+"                            // user and group
            "(\\d+)\\s+"                                      // size
            "((?:\\w+)\\s+(?:\\d+)\\s+(?:\\d+):(?:\\d+))\\s+" // date
            "(\\S+)" );

        QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
        QString line;
        while( !( line = stream.readLine() ).isNull() )
        {

            // parse line
            if( fileRegExp.indexIn(line) < 0 ) continue;

            FileRecord file;
            file.setFile( File( fileRegExp.cap(6) ).addPath( path ) );
            file.addProperty( FileRecordProperties::Permissions, fileRegExp.cap(1) );
            file.addProperty( FileRecordProperties::User, fileRegExp.cap(2) );
            file.addProperty( FileRecordProperties::Group, fileRegExp.cap(3) );
            file.addProperty( FileRecordProperties::Size, fileRegExp.cap(4) );
            file.addProperty( FileRecordProperties::Creation, fileRegExp.cap(5) );

            files << file;

        }

        emit jobFilesAvailable( job, path, files );

    }

    //________________________________________________________________________
    void Interface::_processJobInformation( Job job, bool emitTrace )
    {

        Debug::Throw( "Alien::Interface::_processJobInformation.\n" );

        // regular expression to match data
        // date string is: Tue Nov  1 15:49:06 2011
        QRegExp timeRegExp( "(\\w+)\\s+(\\w+)\\s+(\\d+)\\s+(\\d+):(\\d+):(\\d+)\\s+(\\d+)\\s+" );

        // parent ID regExp  [Master Job is 135064557]
        QRegExp parentRegExp( "\\[Master Job is (\\d+)\\]" );

        // child ID regExp
        QRegExp childRegExp( "Subjob submitted:\\s+(\\d+)" );
        QRegExp resubmissionRegExp( "resubmiting subjob\\s+(\\d+)\\s+\\((\\d+)\\)" );
        QRegExp newJobRegExp( "New job \\(resubmission from (\\d+)\\)" );

        // line type enumeration
        enum LineType
        {
            None = 0,
            Child = 1<<0,
            Parent = 1<<1,
            Resubmission = 1<<2,
            NewJob = 1<<3,
            Starting = 1<<4,
            Running = 1<<5,
            Completed = 1<<6
        };

        // children job set
        JobSet children;

        // id of resubmitted job
        int resubmitted( -1 );

        // loop over lines, and look for trace information
        bool first( true );

        // parse output
        QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
        QString line;
        while( !( line = stream.readLine() ).isNull() )
        {

            // decide line type
            int lineType( None );

            // child job
            Job child;

            if( line.indexOf( "[submit ]" ) >= 0 )
            {

                if( childRegExp.indexIn(line) >= 0 )
                {

                    // child job
                    lineType |= Child;
                    child.assign( Job::Id, childRegExp.cap(1) );

                }

            } else if( line.indexOf( "[state ]" ) >= 0 ) {

                if( parentRegExp.indexIn(line) >= 0 )
                {

                    // parent ID
                    lineType |= Parent;
                    job.setParentId( parentRegExp.cap(1).toInt() );

                } else if( resubmissionRegExp.indexIn( line ) >= 0 ) {

                    // resubmitted job
                    lineType |= Resubmission;
                    child.assign( Job::Id, resubmissionRegExp.cap(2) );

                } else if( newJobRegExp.indexIn(line) >= 0 ) {

                    // new job
                    lineType |= NewJob;
                    resubmitted = newJobRegExp.cap(1).toInt();

                }

                // job time stamps
                if( first ) lineType |= Starting;

                if(
                    line.indexOf( "to SPLIT" ) >= 0 ||
                    line.indexOf( "to ASSIGNED" ) >= 0 ||
                    line.indexOf( "to RUNNING" ) >= 0 ||
                    line.indexOf( "to STARTED" ) >= 0 ) lineType |= Running;

                else if( line.indexOf( "to ERROR" ) >= 0 || line.indexOf( "to DONE" ) >= 0 ) lineType |= Completed;

            }

            // skip if no valid type set
            if( lineType == None ) continue;

            // parse date
            if( timeRegExp.indexIn(line) < 0 ) continue;

            // parse date
            QString date = QString( "%1 %2 %3:%4" ).arg(timeRegExp.cap(2)).arg(timeRegExp.cap(3)).arg(timeRegExp.cap(4)).arg(timeRegExp.cap(5));

            // assign to job
            first = false;
            if( lineType&Starting ) job.assign( Job::SubmissionTime, date );
            if( lineType&Running ) job.assign( Job::StartTime, date );
            if( lineType&Completed ) job.assign( Job::EndTime, date );

            else if( lineType&(Child|Resubmission) )
            {

                // update children list
                auto iter( children.find( child ) );
                if( iter == children.end() ) {

                    // assign submission time and insert in children list
                    child.assign( Job::SubmissionTime, date );
                    children.insert( child );

                } else {

                    // update inserted job
                    const_cast<Job*>( &*iter )->assign( Job::SubmissionTime, date );
                    if( lineType&Resubmission )
                    { const_cast<Job*>( &*iter )->incrementResubmission(); }

                }

            }

        }

        // process job resubmission
        if( resubmitted >= 0 )
        {
            // find previous job from killed jobs list, and reassign parent
            Job sibling = _findKilledJob(  job.get( Job::User ), resubmitted );
            if( sibling.isValid() && sibling.hasParent() )
            { job.setParentId( sibling.parentId() ); }
        }

        // emit signals
        if( emitTrace ) emit jobInformationAvailable( job, stdoutBuffer_.text() );
        else emit jobInformationAvailable( job );

        if( !children.empty() ) emit childrenInformationAvailable( job, children );

        return;
    }

    //________________________________________________________________________
    void Interface::_processJdl( const Job& job )
    {
        QRegExp regExp( "^\\s{8,8}" );

        QTextStream stream( &stdoutBuffer_.text(), QIODevice::ReadOnly );
        QString line;
        QString buffer;
        while( !( line = stream.readLine() ).isNull() )
        {
            if( line == "    [" || line == "    ]" ) continue;
            if( line.indexOf( regExp ) < 0 ) continue;
            line.replace( regExp, "" );
            buffer+= line + "\n";
        }

        // find output directory
        QString outputDir;
        QRegExp outputDirRegExp( "OutputDir\\s+=\\s+\\\"(\\S+)\\\"" );
        if( outputDirRegExp.indexIn( buffer ) >= 0 )
        {
            outputDir = outputDirRegExp.cap(1);
            Debug::Throw() << "Alien::Interface::_processJdl - output directory: " << outputDir << endl;
        }

        // find output files
        QStringList outputFiles;
        QRegExp outputFilesRegExp( "OutputFile\\s+=\\s+\\{\\s+(\\\"\\S+\\\",*\\s+)+\\}" );
        QRegExp fileRegExp( "\\\"(\\S+)\\\"" );
        if( outputFilesRegExp.indexIn( buffer ) >= 0 )
        {
            // capture strings from matched text
            const QString capturedText( outputFilesRegExp.cap(0) );

            int position = 0;
            while(( position = fileRegExp.indexIn( capturedText, position ) ) >= 0 )
            {
                outputFiles.append( fileRegExp.cap(1) );
                position += fileRegExp.matchedLength();
            }

            Debug::Throw() << "Alien::Interface::_processJdl - output files: " << outputFiles.join( " " ) << endl;

        }

        // TODO:
        // - add HTML markups to buffer;
        // - build output files and pass to JdlAvailable
        emit jdlAvailable( job, buffer );
    }

    //___________________________________________________________________
    void Interface::_readStdout()
    {
        process_->setReadChannel( QProcess::StandardOutput );
        QString buffer = QString::fromLocal8Bit( process_->readAll() );

        // replace dummy characters
        const QRegExp regExp( "(\\S){2,2}((\\d{2,2};\\d{2,2})|(0))m" );
        buffer.replace( regExp, "" );

        emit logAvailable( buffer );
        stdoutBuffer_.append( buffer );
    }

    //___________________________________________________________________
    void Interface::_readStderr()
    {
        process_->setReadChannel( QProcess::StandardError );
        const QString buffer = QString::fromLocal8Bit( process_->readAll() );
        emit logAvailable( buffer, Format::Italic );
        stderrBuffer_.append( buffer );
    }

    //_______________________________________
    void Interface::_updateConfiguration()
    {
        Debug::Throw( "Alien::Interface::_updateConfiguration.\n" );

       // get system environment
        systemEnvironment_ = QProcessEnvironment::systemEnvironment();

        // get certificates and check
        const File certificates( XmlOptions::get().raw( "CERTIFICATES_PATH" ) );
        if( certificates.exists() && certificates.isDirectory() )
        {
            systemEnvironment_.insert( "XrdSecSSLCADIR", certificates );
            systemEnvironment_.insert( "X509_CERT_DIR", certificates );
        }

        // get gshell root path from gbbox binary
        const QString gbboxCommand( XmlOptions::get().raw( "GBBOX" ) );
        const QString match( "/bin/gbbox" );
        const int position( gbboxCommand.lastIndexOf( match ) );
        File gShellPath;
        if( position == gbboxCommand.length() - match.length() )
        { gShellPath = File( gbboxCommand.left( position ) ).addTrailingSlash(); }

        // check
        if( gShellPath.exists() && gShellPath.isDirectory() )
        {

            // Store GSHELL_ROOT
            systemEnvironment_.insert( "GSHELL_ROOT", gShellPath );

            // update LD_LIBRARY_PATH if needed
            const QString alienLibPath( gShellPath+"/lib" );
            QString ldPath;
            if( systemEnvironment_.contains( "LD_LIBRARY_PATH" ) )
            { ldPath = systemEnvironment_.value( "LD_LIBRARY_PATH" ); }

            if( ldPath.indexOf( alienLibPath ) < 0 )
            {
                ldPath = ldPath.isEmpty() ? alienLibPath : (alienLibPath + ":" +ldPath);
                systemEnvironment_.insert( "LD_LIBRARY_PATH", ldPath );
            }

        }

    }

    //____________________________________________________________
    void Interface::_resetBuffers()
    {
        Debug::Throw( "Alien::Interface::_resetBuffer.\n" );
        stdoutBuffer_.clear();
        stderrBuffer_.clear();
    }

    //____________________________________________________________
    CustomProcess* Interface::_newProcess()
    {

        Debug::Throw( "Alien::Interface::_newProcess.\n" );

        // create new process and assign customized environment
        process_.reset( new CustomProcess( this ) );
        process_->setProcessEnvironment( systemEnvironment_ );

        connect( process_.get(), SIGNAL(error(QProcess::ProcessError)), SLOT(_error(QProcess::ProcessError)) );
        connect( process_.get(), SIGNAL(finished(int,QProcess::ExitStatus)), SLOT(_commandCompleted(int,QProcess::ExitStatus)) );
        connect( process_.get(), SIGNAL(readyReadStandardOutput()), SLOT(_readStdout()) );
        connect( process_.get(), SIGNAL(readyReadStandardError()), SLOT(_readStderr()) );
        return process_.get();

    }

    //____________________________________________________________
    bool Interface::_isError( const QString& value ) const
    {
        if( value.indexOf( "verify failed" ) >= 0 ) return true;
        else if( value.indexOf( "command failed" ) >= 0 ) return true;
        else if( value.indexOf( "err -1" ) >= 0 ) return true;
        else if( value.indexOf( "err 10" ) >= 0 ) return true;
        else return false;
    }

    //____________________________________________________________
    Job Interface::_findKilledJob( const QString& user, int id ) const
    {
        JobMap::const_iterator iter( killedJobs_.find( user ) );
        if( iter == killedJobs_.end() ) return Job();

        Job local;
        local.assign( Job::Id, QString::number( id ) );
        JobSet::const_iterator jobIter( iter.value().find( local ) );
        return jobIter == iter.value().end() ? Job():*jobIter;
    }

    //_______________________________________
    bool Interface::_ignoreError( const Command& command, const QString& message ) const
    {

        // check connection state
        if( !( connected_ || command.type() == Command::Type::Connect ) ) return false;

        return false;

    }
}
