#ifndef AlienInterface_h
#define AlienInterface_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AlienCommand.h"
#include "AlienCommandList.h"
#include "CustomProcess.h"
#include "Debug.h"
#include "File.h"
#include "FileRecord.h"
#include "Job.h"
#include "JobSet.h"
#include "MessageBuffer.h"
#include "TextFormat.h"
#include "UserSet.h"

#include <QObject>
#include <QProcess>
#include <QProcessEnvironment>
#include <QString>
#include <QHash>

namespace Alien
{
    class Interface:public QObject, private Base::Counter<Interface>
    {

        //* Qt meta object declaration
        Q_OBJECT

        public:

        //* constructor
        explicit Interface( QObject* = nullptr );

        //* find certificates
        void findCertificates();

        //* clear failed commands
        void clearFailedCommands()
        { failedCommands_.clear(); }

        //* hide killed jobs
        void setHideKilledJobs( bool value )
        { hideKilledJobs_ = value; }

        //* true if connected
        bool isConnected() const
        { return connected_; }

        Q_SIGNALS:

        //* connection lost
        void connectionLost( const QString& = QString() );

        //* connection state changed
        void connectionStateChanged( bool );

        //* default user is available
        void defaultUserAvailable( QString );

        //* user list is available
        void userListAvailable( UserSet );

        //* master list is available
        void masterListAvailable( QString, const JobSet& );

        //* children list is available
        void childrenListAvailable( const Job&, const JobSet& );

        //* job information is available
        void jobInformationAvailable( const Job&, const QString& = QString() );

        //* children information is available
        void childrenInformationAvailable( const Job&, const JobSet& );

        //* job output files
        void jobFilesAvailable( const Job&, const QString&, const FileRecord::List& );

        //* job JDL is available
        void jdlAvailable( const Job&, const QString& );

        //* command added
        void commandAdded( const Alien::Command& );

        //* commands added
        void commandsAdded( const Alien::Command::List& );

        //* command started
        void commandStarted( const Alien::Command& );

        //* command completed
        void commandCompleted( const Alien::Command& );

        //* commands failed
        void commandsFailed( const Alien::Command::List& );

        //* emitted when busy state is changed
        void busyStateChanged( bool );

        //* emitted when busy processing commands
        void busy();

        //* emitted when there is no more commands in pile
        void ready();

        //* logAvailable
        void logAvailable( const QString&, Format::TextFormatFlags = 0 );

        //* message
        void messageAvailable( const QString& );

        public Q_SLOTS:

        //* add command
        void addCommand( const Alien::Command& );

        //* add commands
        void addCommands( const Alien::Command::List& );

        //* abort all commands
        void abortAll();

        //* abort selected commands
        void abortCommands( const Alien::Command::List& );

        //* toggle pause
        void togglePause( bool value )
        {
            paused_ = value;
            if( !paused_ ) _runCommand();
        }

        //* update set of active users
        void updateActiveUsers( UserSet users )
        { activeUsers_ = users; }

        //* clear queue
        void clearQueue( const QString&, const Job::IdSet& );

        private Q_SLOTS:

        //* run next command in pile
        bool _runCommand();

        //* take action when current command completed
        void _commandCompleted( int, QProcess::ExitStatus );

        //* error from process
        void _error( QProcess::ProcessError );

        //* read on-fly from process
        void _readStdout();

        //* read on-fly from process
        void _readStderr();

        //* update configuration
        void _updateConfiguration();

        private:

        //* check connection
        bool _checkConnection();

        //* process proxy
        void _processProxyInfo();

        //* process default user
        void _processDefaultUser();

        //* process user list
        void _processUserList();

        //* process master job list
        void _processMasterList( const QStringList& );

        //* process children job list
        void _processChildrenList( const Job& );

        //* job output files
        void _processJobFiles( const Job&, const File& );

        //* process job set
        void _processJobInformation( Job, bool );

        //* process job set
        void _processJdl( const Job& );

        //* reset buffers
        void _resetBuffers();

        //* attempt at identifying error lines
        bool _isError( const QString& ) const;

        //* find killed job in stored map
        /** returns invalid job if not found */
        Job _findKilledJob( const QString&, int ) const;

        //* true if given error is to be ignored
        bool _ignoreError( const Command&, const QString& ) const;

        //* create new process to run command
        CustomProcess* _newProcess();

        //* true if paused
        bool paused_ = false;

        //* hide killed jobs
        bool hideKilledJobs_ = false;

        //* custom process
        CustomProcess::Pointer process_;

        //* stdout buffer
        MessageBuffer stdoutBuffer_;

        //* stderr buffer
        MessageBuffer stderrBuffer_;

        //* map command type to name
        CommandList commands_;

        //* failed commands
        Command::List failedCommands_;

        //* set of active users
        UserSet activeUsers_;

        //* map users to (killed) job set
        using JobMap = QHash< QString, JobSet >;
        JobMap killedJobs_;

        //* system environment
        QProcessEnvironment systemEnvironment_;

        //* true if connection was valid at least once
        bool wasConnected_ = false;

        //* connected flag
        bool connected_ = true;

        //* ignore next error (due to abort)
        bool ignoreErrors_ = false;

    };

}
#endif
