/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "JdlSyntaxHighlighter.h"
#include "CppUtil.h"
#include "Debug.h"
#include "TextBlockRange.h"
#include "TextEditor.h"

#include <QPalette>
#include <QTextCursor>

//__________________________________________________________
JdlSyntaxHighlighter::JdlSyntaxHighlighter( QTextDocument* parent ):
    QSyntaxHighlighter( parent ),
    Counter( "JdlSyntaxHighlighter" )
{ Debug::Throw( "JdlSyntaxHighlighter::JdlSyntaxHighlighter.\n" ); }

//__________________________________________________________
void JdlSyntaxHighlighter::updateLinks()
{
    if( !document() ) return;

    // regular expressions
    auto regExps = Base::makeT< QList<QRegExp> >( {
        QRegExp("\"(?:LF\\:)?(/.+)(?:\\,|\")"),
        QRegExp("(?:(?:OutputDir)|(?:Executable)|(?:Validationcommand))\\s*=\\s*\"(.+)\";")
    });

    regExps[0].setMinimal( true );

    // loop over document blocks
    for( const auto& block:TextBlockRange( document() ) )
    {

        // create cursor
        QTextCursor cursor( block );

        // loop over regular expressions
        for( const auto& regExp:regExps )
        {
            const auto text( block.text() );
            for( int position = 0; ( position = text.indexOf( regExp, position ) ) >= 0; position += regExp.matchedLength() )
            {

                // find emoticon matching regexp
                const int localPosition = regExp.pos(1);
                const QString link( regExp.cap(1) );
                const int length( link.size() );

                // set cursor at the right position
                cursor.movePosition( QTextCursor::StartOfBlock );
                cursor.movePosition( QTextCursor::Right, QTextCursor::MoveAnchor, localPosition );
                cursor.movePosition( QTextCursor::Right, QTextCursor::KeepAnchor, length );

                _addLink( cursor, link );

            }

        }

    }

}

//__________________________________________________________
void JdlSyntaxHighlighter::highlightBlock( const QString & text )
{

    enum { IOState, ListState };

    if( previousBlockState() == IOState || previousBlockState() == ListState )
    {

        // propagate to next block
        if( text.indexOf( QRegExp( "\\}" ) ) < 0 ) setCurrentBlockState( previousBlockState() );

    } else {

        // bold prefix
        QRegExp regExp;
        regExp.setPattern( "\\s*(\\S+)\\s*=" );
        if( text.indexOf( regExp ) >= 0 )
        {
            const int position = regExp.pos(1);
            const int length = regExp.cap(1).size();
            _addBold( position, length );
        }

        // list of files
        if( text.indexOf( QRegExp( "((OutputFile)|(InputBox)|(InputData)|(InputFile))\\s*=" ) ) >= 0 ) setCurrentBlockState( IOState );
        else if(text.indexOf( QRegExp( "\\{" ) ) >= 0 ) setCurrentBlockState( ListState );
    }

}

//__________________________________________________________
void JdlSyntaxHighlighter::_addBold( int position, int length )
{
    QTextCharFormat format;
    format.setFontWeight( QFont::DemiBold );
    setFormat( position, length, format );
}

//__________________________________________________________
void JdlSyntaxHighlighter::_addLink( QTextCursor cursor, const QString& link )
{
    // create relevant char format
    QTextCharFormat format;
    format.setFontUnderline( true );
    format.setForeground( QPalette().color( QPalette::Link ) );
    format.setAnchorHref( QString( "alien://%1" ).arg( link ) );
    format.setAnchor( true );

    // apply to selection
    cursor.mergeCharFormat( format );
}
