/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Job.h"

#include "CppUtil.h"
#include "File.h"
#include "JobSet.h"
#include "XmlDef.h"

#include <QRegExp>

//____________________________________________________________
const Job::FieldVector& Job::_fieldNames()
{
    static const auto fieldNames = Base::makeT<FieldVector>
    ({
        QObject::tr( "Id" ),
        QObject::tr( "Owner " ),
        QObject::tr( "Status" ),
        QObject::tr( "Command " ),
        QObject::tr( "Resubmitted " ),
        QObject::tr( "Submitted" ),
        QObject::tr( "Started" ),
        QObject::tr( "Completed" )
    });

    return fieldNames;
}

//____________________________________________________________
const Job::FieldVector& Job::_xmlNames()
{
    static const auto xmlNames = Base::makeT<FieldVector>({ "i", "o", "s", "t", "r", "su", "st", "co" });
    return xmlNames;
}

//____________________________________________________________
const Job::FieldMap& Job::_xmlMap()
{
    auto makeMap = []( const Job::FieldVector& names )
    {
        Job::FieldMap out;
        for( int i=0; i < names.size(); ++i )
        { out.insert( names[i], i ); }
        return out;
    };

    static const FieldMap xmlMap = makeMap( _xmlNames() );
    return xmlMap;
}

//____________________________________________________________
const Job::ConversionMap &Job::_conversions()
{
    static const auto conversions = Base::makeT<ConversionMap>( {
        { "WAITING", "W" },
        { "INSERTING", "I" },
        { "RUNNING", "R" },
        { "STARTED", "ST" },
        { "SAVING", "SV" },
        { "SAVED", "SVD" },
        { "DONE", "D" },
        { "ERROR_V", "EV" },
        { "ERROR_VN", "EV" },
        { "ERROR_SV", "ESV" },
        { "ERROR_E", "EE" },
        { "ERROR_RE", "ERE" },
        { "ZOMBIE", "Z" }
    });

    return conversions;
}

//____________________________________________________________
Job::Job( const QDomElement& element ):
    Counter( "Job" ),
    parentId_( -1 ),
    fields_( nFields )
{

    Debug::Throw( "Job::Job (dom).\n" );

    // parse attributes
    const auto attributes( element.attributes() );
    for( int i=0; i<attributes.count(); i++ )
    {
        const auto attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;
        if( attribute.name() == Xml::Parent ) setParentId( attribute.value().toInt() );
    }

    // parse children elements
    for( auto&& childNode = element.firstChild(); !childNode.isNull(); childNode = childNode.nextSibling() )
    {
        const auto childElement = childNode.toElement();
        const auto tagName( childElement.tagName() );
        if( tagName == Xml::Field ) _loadXmlField( childElement );
    }

}

//____________________________________________________________
Job::Job( const QString& buffer, Type type ):
    Counter( "Job" ),
    parentId_( -1 ),
    fields_( nFields )
{
    // split buffer
    const auto fields( buffer.split( QRegExp("\\s+") ) );

    // check field size
    if( fields.size() < 3 ) return;

    switch( type )
    {
        case ParseMaster:
        {
            assign( User, fields[0] );
            assign( Id, fields[1] );
            assign( Status, fields[2] );

            if( fields.size() > 3 )
            {
                assign( Command, File( fields[3] ).localName() );
                Debug::Throw() << "Job::Job - command (ParseMaster): " << get( Command ) << endl;
            }
            break;
        }

        case ParseChildren:
        {
            assign( Id, fields[0] );

            // need to convert status to matching short string
            // todo: cood use conversion map for speeding up
            const QString status( fields[1] );
            auto iter = _conversions().find( status );
            assign( Status, iter == _conversions().end() ? status:iter.value() );

            /*
            there are two possible formats for gbbox ps. One for which command name is at index 2, and one
            for which it is at index 3. Need to check both
            */
            if( fields[2].startsWith( "/alice/cern.ch/" ) ) assign( Command, File( fields[2] ).localName() );
            else if( fields.size() > 3 ) assign( Command, File( fields[3] ).localName() );

            break;
        }

        default: break;

    }


    return;
}

//____________________________________________________________
QDomElement Job::domElement( QDomDocument& document ) const
{

    QDomElement out = document.createElement( Xml::Job );
    out.setAttribute( Xml::Parent, QString::number( parentId() ) );

    // add fields
    for( int index=0; index<fields_.size(); ++index )
    {

        if( !has( index ) ) continue;
        auto field = out.appendChild( document.createElement( Xml::Field ) ).toElement();
        field.setAttribute( Xml::FieldName, _xmlNames()[index] );
        field.appendChild( document.createTextNode( fields_[index] ) );

    }

    return out;

}

//____________________________________________________________
QString Job::status() const
{

    if( isFailed() ) return QString( QObject::tr( "Failed (%1)" ) ).arg( error() );
    else {

        QString st( get( Status ) );
        if( id() == 0 ) return "";
        else if( st.isEmpty() ) return QObject::tr( "Unknown" );
        else if( st == "I" ) return QObject::tr( "Inserting" );
        else if( st == "SP" || st == "WS" ) return QObject::tr( "Splitting" );
        else if( st == "R" || st == "RS" ) return QObject::tr( "Running" );
        else if( st == "D" ) return QObject::tr( "Completed" );
        else if( st == "DONE_WARN" ) return QObject::tr( "Completed (warning)" );
        else if( st == "SAVED_WARN" ) return QObject::tr( "Saved (warning)" );
        else if( st == "W" ) return QObject::tr( "Waiting" );
        else if( st == "OVER_WAITING" ) return QObject::tr( "Over waiting" );
        else if( st == "ASSIGNED" ) return QObject::tr( "Assigned" );
        else if( st == "S" ) return QObject::tr( "Split" );
        else if( st == "SV" ) return QObject::tr( "Saving" );
        else if( st == "SVD" ) return QObject::tr( "Saved" );
        else if( st == "ST" ) return QObject::tr( "Starting" );
        else if( st == "K" ) return QObject::tr( "Killed" );
        else if( st == "Z" ) return QObject::tr( "Zombie" );
        else return st;
    }
}

//____________________________________________________________
QString Job::error() const
{

    QString st( get( Status ) );
    if( st.isEmpty() ) return QString();
    else if( st == "EI" ) return QObject::tr( "insertion" );
    else if( st == "EA" ) return QObject::tr( "assignment" );
    else if( st == "ES" ) return QObject::tr( "queuing" );
    else if( st == "EE" ) return QObject::tr( "execution" );
    else if( st == "ER" ) return QObject::tr( "running" );
    else if( st == "EV" ) return QObject::tr( "validation" );
    else if( st == "EIB" ) return QObject::tr( "input box" );
    else if( st == "ESV" ) return QObject::tr( "saving" );
    else if( st == "ESP" ) return QObject::tr( "splitting" );
    else if( st == "ERE" ) return QObject::tr( "reading" );
    else if( st == "EXPIRED" ) return QObject::tr( "expired" );
    else return st;

}

//____________________________________________________________
QString Job::get( int index ) const
{
    if( index < nFields ) return fields_[index];
    else return QString::number( childrenStatistics_.get( index - nFields ) );
}

//____________________________________________________________
Job& Job::assign( int index, const QString& value )
{
    fields_[index] = value;
    switch( index )
    {
        case SubmissionTime: submissionTime_ = _time( value ); break;
        case StartTime: startTime_ = _time( value ); break;
        case EndTime: endTime_ = _time( value ); break;
        default: break;
    }
    return *this;
}

//____________________________________________________________
void Job::incrementResubmission()
{
    const int resubmission( has( Resubmitted ) ? get<int>( Resubmitted ):0 );
    assign<int>( Resubmitted, resubmission+1 );
}

//____________________________________________________________
bool Job::isInserting() const
{ return get( Status ) == "I"; }

//____________________________________________________________
bool Job::isRunning() const
{
    const QString status( get( Status ) );
    return
        (isMasterJob() && status == "I") ||
        status == "SP" || // splitting
        status == "R" ||
        status == "RS" ||
        status == "S" ||
        status == "ST" ||
        status == "SV" ||
        status == "SVD" ||
        status == "SAVED_WARN";
}

//____________________________________________________________
bool Job::isCompleted() const
{
    const QString status( get( Status ) );
    return
        status == "D" ||
        status == "DONE_WARN";
}

//____________________________________________________________
bool Job::isPending() const
{
    const QString status( get( Status ) );
    return
        (hasParent() && status == "I" ) ||
        status == "W" ||
        status == "OVER_WAITING" ||
        status == "ASSIGNED";
}

//____________________________________________________________
bool Job::isFailed() const
{
    const QString status( get( Status ) );
    return status.startsWith( 'E' ) || status == "Z";
}

//____________________________________________________________
bool Job::isKilled() const
{ return get( Status ) == "K"; }

//____________________________________________________________
bool Job::mightHaveChildren() const
{

    // list of status values corresponding to master jobs that have no children
    using StatusSet = QSet<QString>;
    static const auto statusSet = Base::makeT<StatusSet>( {
        "W", "R", "I",
        "SV", "SVD",
        "K",
        "EE", "ER", "EV",
        "ESV", "EIB",
        "Z", "EXPIRED"
    } );

    // true if status is not in above list
    return !statusSet.contains( get( Status ) );

}

//____________________________________________________________
QString Job::resubmitted() const
{
    if( !has( Job::Resubmitted ) ) return QString();
    else {

        const int count( get<int>( Job::Resubmitted ) );
        return count == 1 ? QString( QObject::tr( "%1 time" ) ).arg( count ) : QString( QObject::tr( "%1 times" ) ).arg( count );

    }

}

//____________________________________________________________
void Job::_loadXmlField( const QDomElement& element )
{
    Debug::Throw( "Job::_loadXmlField.\n" );
    const auto attributes( element.attributes() );
    for( int i=0; i<attributes.count(); i++ )
    {
        const auto attribute( attributes.item( i ).toAttr() );
        if( attribute.isNull() ) continue;
        const auto name( attribute.name() );
        const auto value( attribute.value() );

        if( name != Xml::FieldName ) continue;

        int id( 0 );
        if( _xmlToId( value, id ) )
        {
            assign( id, element.text() );
            break;
        }
    }
}

//____________________________________________________________
TimeStamp Job::_time( const QString& value )
{

    QRegExp regExp( "(\\w+)\\s+(\\d+)\\s+(\\d+):(\\d+)" );
    const int index( regExp.indexIn( value ) );
    if( index < 0 || regExp.captureCount() < 4 ) return TimeStamp();

    const auto captures( regExp.capturedTexts() );
    TimeStamp out( TimeStamp::now() );
    out.setMonth( captures[1] )
        .setDay( captures[2].toInt() )
        .setHours( captures[3].toInt() )
        .setMinutes( captures[4].toInt() ).makeTime();

    return out;
}

//____________________________________________________________
Job::IdSet Job::idSet( const JobSet& jobs )
{
    Job::IdSet out;
    for( const auto& job:jobs )
    { if( job.id() >= 0 ) out.insert( job.id() ); }

    return out;
}

//____________________________________________________________
Job::IdSet Job::idSet( const Job::List& jobs )
{
    Job::IdSet out;
    for( const auto& job:jobs )
    { if( job.id() >= 0 ) out.insert( job.id() ); }

    return out;
}
