#ifndef Job_h
#define Job_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "Debug.h"
#include "Functors.h"
#include "JobStatistics.h"
#include "TimeStamp.h"
#include "QOrderedSet.h"

#include <QDomDocument>
#include <QDomElement>
#include <QString>
#include <QHash>
#include <QVector>
#include <QHash>

#include <cmath>

//* forward declaration
class JobSet;

//* job
class Job final: private Base::Counter<Job>
{

    public:

    //* allows fast, keyed access to any data field
    enum FieldType
    {
        Id,
        User,
        Status,
        Command,
        Resubmitted,
        SubmissionTime,
        StartTime,
        EndTime,
        nFields
    };

    //* get field name
    static const QString& fieldName( int index )
    { return _fieldNames()[index]; }

    //* empty constructor
    explicit Job():
        Counter( "Job" ),
        parentId_( -1 ),
        fields_( nFields )
    {}

    //: constructor from dom
    explicit Job( const QDomElement& );

    //* parsing type
    enum Type
    {
        ParseMaster,
        ParseChildren
    };

    //* constructor from string to browse
    explicit Job( const QString&, Type );

    //* dom element
    QDomElement domElement( QDomDocument& ) const;

    //*@name modifiers
    //@{

    //* parent id
    Job& setParentId( int value )
    {
        parentId_ = value;
        return *this;
    }

    //* children statistics
    Job& setChildrenStatistics( const JobStatistics& value )
    {
        childrenStatistics_ = value;
        return *this;
    }

    //* set field
    Job& assign( int, const QString& );

    //* set field
    template <typename T>
        Job& assign( int index, const T& value )
    {
        QString stringValue;
        {
            QTextStream stream( &stringValue, QIODevice::WriteOnly );
            stream << value;
        }
        return assign( index, stringValue );
    }

    //* increment resubmission
    void incrementResubmission();

    //@}

    //*@name accessors
    //@{

    //* valid
    bool isValid() const
    {
        bool error( false );
        get<int>(Id, &error );
        return (!error && has( User ));
    }

    //* id
    int id() const
    { return has(Id) ? std::abs(get<int>(Id)):-1; }

    //* parent id
    int parentId() const
    { return parentId_; }

    //* returns true job is top level
    bool isMasterJob() const
    { return !hasParent(); }

    //* returns true if job has a parent
    bool hasParent() const
    { return parentId() > 0; }

    //* returns true if has children
    bool hasChildren() const
    { return childrenStatistics_.get( JobStatistics::Total ) > 0; }

    //* children statistics
    const JobStatistics& childrenStatistics() const
    { return childrenStatistics_; }

    //* returns true if this job is a child of argument
    bool isChild( const Job& job ) const
    { return parentId() == job.id(); }

    //* true if inserting
    bool isInserting() const;

    //* true if running
    bool isRunning() const;

    //* true if completed
    bool isCompleted() const;

    //* true if pending
    bool isPending() const;

    //* true if has error
    bool isFailed() const;

    //* true if job is killed
    bool isKilled() const;

    //* true if job might have children (based on status)
    bool mightHaveChildren() const;

    //* return status
    QString status() const;

    //* return error code
    QString error() const;

    //* const field accessor
    QString get( int index ) const;

    //* field accessor (cast to anything)
    template <typename T>
        T get( int index, bool* error = 0 ) const
    {
        T out;
        QString value( get( index ) );
        QTextStream in( &value, QIODevice::ReadOnly );
        in >> out;
        if( in.status() != QTextStream::Ok )
        {
            if( error ) *error = true;
            return T();
        } else return out;
    }

    //* Check if field makes sence
    bool has( int index ) const
    { return get(index).size(); }

    //* resubmission count
    QString resubmitted() const;

    //* times
    const TimeStamp& submissionTime() const
    { return submissionTime_; }

    //* times
    const TimeStamp& startTime() const
    { return startTime_; }

    //* times
    const TimeStamp& endTime() const
    { return endTime_; }

    //@}

    //* used to count total number of jobs
    class TotalFTor
    {
        public:

        //* predicate
        bool operator() (const Job& job ) const
        {  return !( job.isMasterJob() && job.hasChildren() ); }

    };

    //* used to select pending jobs
    class PendingFTor
    {
        public:

        //* constructor
        explicit PendingFTor()
        {}

        //* predicate
        bool operator() (const Job& job ) const
        {  return job.isPending() && !job.hasChildren(); }

    };

    //* used to select running jobs
    class RunningFTor
    {
        public:

        //* constructor
        explicit RunningFTor()
        {}

        //* predicate
        bool operator() (const Job& job ) const
        {  return job.isRunning() && !job.hasChildren(); }

    };

    //* used to select completed jobs
    class CompletedFTor
    {
        public:

        //* constructor
        explicit CompletedFTor()
        {}

        //* predicate
        bool operator() (const Job& job ) const
        {  return job.isCompleted() && !job.hasChildren(); }

    };

    //* used to select failed jobs
    class FailedFTor
    {
        public:

        //* constructor
        explicit FailedFTor()
        {}

        //* predicate
        bool operator() (const Job& job ) const
        {  return job.isFailed() && !job.hasChildren(); }

    };

    //* used to select jobs with given Id
    using SameIdFTor = Base::Functor::Unary<Job, int, &Job::id>;

    //* typedef for list of jobs
    using id_t = int;
    using List = QList<Job>;
    using IdSet = QOrderedSet<id_t>;

    static IdSet idSet( const JobSet& );
    static IdSet idSet( const Job::List& );

    protected:

    //* xml name
    static QString _xmlName( int id )
    { return _xmlNames()[id]; }

    //* set field id from xml field name return false if not found
    static bool _xmlToId( const QString& name, int& id )
    {
        auto&& iter = _xmlMap().find( name );
        if( iter == _xmlMap().end() ) return false;
        id = iter.value();
        return true;
    }

    //* field names
    using FieldVector = QVector<QString>;
    static const FieldVector& _fieldNames();
    static const FieldVector& _xmlNames();

    //* xml map
    using FieldMap = QHash<QString, int>;
    static const FieldMap& _xmlMap();

    //* status conversions
    using ConversionMap = QHash<QString,QString>;
    static const ConversionMap& _conversions();

    //* load field from DomElement
    void _loadXmlField( const QDomElement& element );

    // convert string to time
    TimeStamp _time( const QString& );

    private:

    //* parent id
    /** it is kept outside of fields in order not to appear in list */
    int parentId_ = 0;

    //* fields
    FieldVector fields_;

    //* children statistics
    JobStatistics childrenStatistics_;

    // timestamps
    TimeStamp submissionTime_;
    TimeStamp startTime_;
    TimeStamp endTime_;
};

//* lower than operator
inline bool operator < ( const Job& first, const Job& second )
{ return first.id() < second.id(); }

//* equal to operator
inline bool operator == (const Job& first, const Job& second )
{ return first.id() == second.id(); }

//* hash
inline uint qHash( const Job& job )
{ return qHash( job.id() ); }

#endif
