/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "JobModel.h"

#include "Color.h"
#include "CustomPixmap.h"
#include "Operators.h"
#include "Singleton.h"
#include "XmlOptions.h"

#include <QPainter>

//_______________________________________________________________
JobModel::JobModel( QObject* parent ):
    TreeModel( parent ),
    Counter( "JobModel" ),
    today_( TimeStamp::now() )
{
    Debug::Throw( "JobModel::JobModel.\n" );
    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();
}

//__________________________________________________________________
QVariant JobModel::data( const QModelIndex& index, int role ) const
{

    // check index, role and column
    if( !index.isValid() ) return QVariant();

    // return text associated to file and column
    switch( role )
    {
        case Qt::DisplayRole:
        {

            // retrieve associated file info
            const Job& job( _find( index.internalId() ).get() );
            switch( index.column() )
            {

                case Job::SubmissionTime:
                return _timeString( job.submissionTime() );

                case Job::StartTime:
                return _timeString( job.startTime() );

                case Job::EndTime:
                return _timeString( job.endTime() );

                default:
                if( index.column() == Job::Id )
                {
                    const int value( job.id() );
                    if( value ) return value;
                    else return "Orphans";
                } else if( index.column() == Job::Resubmitted ) {

                    return job.resubmitted();

                } else if( index.column() == Job::Status ) return job.status();
                else if( index.column() < Job::nFields ) return job.get( index.column() );
                else if( index.column() - Job::nFields < JobStatistics::nFields && job.hasChildren() )
                {
                    const int value = job.childrenStatistics().get( JobStatistics::FieldType(index.column() - Job::nFields) );
                    if( value ) return value;
                    else return "";
                }
                break;
            }

        }

        case Qt::DecorationRole:
        {
            if( index.column() == Job::Status )
            {

                // retrieve associated file info
                const Job& job( _find( index.internalId() ).get() );
                if( !job.id() ) return QVariant();
                else if( job.isCompleted() ) return completedJobIcon_;
                else if( job.isRunning() ) return runningJobIcon_;
                else if( job.isFailed() ) return failedJobIcon_;
                else return pendingJobIcon_;
            }
            break;

        }

        case Qt::TextAlignmentRole:
        {
            if( index.column() < Job::nFields )
            {
                switch( index.column() )
                {
                    case Job::SubmissionTime:
                    case Job::StartTime:
                    case Job::EndTime:
                    return (int)(Qt::AlignCenter);

                    case Job::User:
                    case Job::Command:
                    case Job::Resubmitted:
                    return (int)(Qt::AlignRight|Qt::AlignVCenter);

                    default: break;
                }

            } else {

                return (int)(Qt::AlignCenter);

            }
            break;

        }

        default: break;
    }

    return QVariant();

}

//__________________________________________________________________
QVariant JobModel::headerData(int section, Qt::Orientation orientation, int role) const
{

    switch( role )
    {
        case Qt::DisplayRole:
        {
            if( section < int(Job::nFields) ) return Job::fieldName( section );
            else if( section - Job::nFields < JobStatistics::nFields ) return JobStatistics::fieldName( section - Job::nFields );
            break;
        }

        case Qt::TextAlignmentRole:
        {
            if( section < Job::nFields )
            {
                switch( section )
                {
                    case Job::SubmissionTime:
                    case Job::StartTime:
                    case Job::EndTime:
                    return (int)(Qt::AlignCenter);

                    case Job::User:
                    case Job::Command:
                    case Job::Resubmitted:
                    return (int)(Qt::AlignRight|Qt::AlignVCenter);

                    default: break;
                }

            } else {

                return (int)(Qt::AlignCenter);

            }
            break;
        }

        default: break;

    }

    // return empty
    return QVariant();

}

//________________________________________________________
bool JobModel::SortFTor::operator () ( Job first, Job second ) const
{

    if( order_ == Qt::DescendingOrder ) std::swap( first, second );
    if( type_ < Job::nFields )
    {

        switch( type_ )
        {
            case Job::Id: return first < second;

            case Job::Status:
            {
                if( first.get( Job::Status ) == second.get( Job::Status ) )
                {

                    return first < second;

                } else if( first.isFailed() ) {

                    if( second.isFailed() ) return first.get( Job::Status ) < second.get( Job::Status );
                    else return true;

                } else if( second.isFailed() ) {

                    return false;

                } else if( first.isRunning() ) {

                    if( second.isRunning() ) return first.get( Job::Status ) < second.get( Job::Status );
                    else return true;

                } else if( second.isRunning() ) {

                    return false;

                } else if( first.isCompleted() ) {

                    if( second.isCompleted() ) return first.get( Job::Status ) < second.get( Job::Status );
                    else return false;

                } else if( second.isCompleted() ) {

                    return true;

                } else {

                    return first.get( Job::Status ) < second.get( Job::Status );

                }

            }

            case Job::SubmissionTime: return first.submissionTime() != second.submissionTime() ?
                first.submissionTime() < second.submissionTime() :
                first < second;

            case Job::StartTime: return first.startTime() != second.startTime() ?
                first.startTime() < second.startTime() :
                first < second;

            case Job::EndTime: return first.endTime() != second.endTime() ?
                first.endTime() < second.endTime() :
                first < second;

            default: return first.get( type_ ) != second.get( type_ ) ?
                first.get( type_ ) < second.get( type_ ) :
                first < second;

            }

    } else if( type_-Job::nFields < JobStatistics::nFields ) {

        const int type( type_-Job::nFields );
        return first.childrenStatistics().get(type) != second.childrenStatistics().get(type) ?
            first.childrenStatistics().get(type) < second.childrenStatistics().get(type) :
            first < second;

    } else return first < second;

}

//__________________________________________________________
void JobModel::_updateConfiguration()
{

    Debug::Throw( "JobModel::_updateConfiguration.\n" );
    runningJobIcon_ = _createIcon( Base::Color( XmlOptions::get().get<Base::Color>("RUNNING_JOB_COLOR") ) );
    completedJobIcon_ = _createIcon( Base::Color( XmlOptions::get().get<Base::Color>("COMPLETED_JOB_COLOR") ) );
    failedJobIcon_ = _createIcon( Base::Color( XmlOptions::get().get<Base::Color>("FAILED_JOB_COLOR") ) );

    // empty icon for pending jobs
    int iconSize = qMax<int>( XmlOptions::get().get<int>( "LIST_ICON_SIZE" ), 16 );
    pendingJobIcon_ = QIcon( CustomPixmap( QSize( iconSize, iconSize ), CustomPixmap::Flag::Transparent ) );

}

//____________________________________________________________
QString JobModel::_timeString( const TimeStamp& time ) const
{
    if( !time.isValid() ) return QString();
    else if( time.isSameDay( today_ ) ) return time.toString( TimeStamp::Format::Time );
    else if( time.isSameYear( today_ ) ) return time.toString( "MMM d hh:mm" );
    else return time.toString( "MMM d yyy hh:mm" );
}

//____________________________________________________________
QIcon JobModel::_createIcon( const QColor& color ) const
{

    Debug::Throw( "JobModel::_createIcon" );

    int iconSize = qMax<int>( XmlOptions::get().get<int>( "LIST_ICON_SIZE" ), 16 );
    double pixmapSize = 0.75* XmlOptions::get().get<double>( "LIST_ICON_SIZE" );
    double offset = 0.5*( iconSize - pixmapSize );

    CustomPixmap pixmap( QSize( iconSize, iconSize ), CustomPixmap::Flag::All );

    if( color.isValid() )
    {

        QPainter painter( &pixmap );
        painter.setRenderHints(QPainter::Antialiasing );
        painter.setPen( Qt::NoPen );

        QRectF rect( QPointF( offset, offset ), QSizeF( pixmapSize, pixmapSize ) );
        QLinearGradient gradient(rect.topLeft(), rect.bottomRight());
        gradient.setColorAt(0, color );
        gradient.setColorAt(1, color.light(135) );
        painter.setBrush( gradient );
        painter.drawEllipse( rect );
        painter.end();

    }

    return QIcon( pixmap );

}
