#ifndef MainWindow_h
#define MainWindow_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AlienCommandModel.h"
#include "BaseMainWindow.h"
#include "BusyWidget.h"
#include "Counter.h"
#include "CustomToolBar.h"
#include "Debug.h"
#include "Job.h"
#include "AlienCommand.h"
#include "Key.h"
#include "UserSet.h"

#include <QCloseEvent>
#include <QTabWidget>

class JobManagerFrame;
class JobSet;
class LogDialog;
class ProgressStatusBar;
class SummaryFrame;
class UserSelectionFrame;

namespace Alien
{
    class Interface;
}

class MainWindow: public BaseMainWindow, public Base::Key
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    explicit MainWindow( QWidget* = nullptr );

    //* user
    UserSelectionFrame& userSelectionFrame() const
    { return *userSelectionFrame_; }

    //* toolbar
    CustomToolBar& userToolBar()
    { return *userToolBar_; }

    //* toolbar
    CustomToolBar& toolBar()
    { return *toolbar_; }

    //* status bar
    ProgressStatusBar& statusBar() const
    { return *statusBar_; }

    //* install frames
    void installFrames();

    //*@name actions
    //@{

    // connect
    QAction& connectAction() const
    { return *connectAction_; }

    //* command list action
    QAction& commandListAction() const
    { return *commandListAction_; }

    //* log action
    QAction& logAction() const
    { return *logAction_; }

    //* log action
    QAction& toggleLogAction() const
    { return *toggleLogAction_; }

    //* hide killed jobs
    QAction& hideKilledJobsAction() const
    { return *hideKilledJobsAction_; }

    //* pause action
    QAction& pauseAction() const
    { return *pauseAction_; }

    //* default user action
    QAction& defaultUserAction() const
    { return *defaultUserAction_; }

    //* user list action
    QAction& userListAction() const
    { return *userListAction_; }

    //* toggle summary
    QAction& toggleSummaryAction() const
    { return *toggleSummaryAction_; }

    //* update
    QAction& updateAction() const
    { return *updateAction_; }

    //* kill
    QAction& killAction() const
    { return *killAction_; }

    //* add frame
    QAction& addFrameAction() const
    { return *addFrameAction_; }

    //* remove current frame
    QAction& removeFrameAction() const
    { return *removeFrameAction_; }

    //* resubmit
    QAction& resubmitAction() const
    { return *resubmitAction_; }

    //* retrace
    QAction& traceAction() const
    { return *traceAction_; }

    //* clear
    QAction& clearAction() const
    { return *clearAction_; }

    //* abort
    QAction& abortAction() const
    { return *abortAction_; }

    //* job dialogs
    QAction& closeJobDialogsAction() const
    { return *closeJobDialogsAction_; }

    //@}

    Q_SIGNALS:

    //* request command
    // TODO: remove
    void commandRequested( const Alien::Command& );

    //* request commands
    // TODO: remove
    void commandsRequested( const Alien::Command::List& );

    //* user list ready
    void userListAvailable(QStringList);

    //* active users have changed
    void activeUsersChanged( UserSet );

    //* scratch file
    void scratchFileCreated( File );

    protected:

    //* timer event
    void timerEvent( QTimerEvent* ) override;

    //* close window event handler
    void closeEvent( QCloseEvent* ) override;

    private Q_SLOTS:

    //* check proxy
    void _checkProxy();

    //* connect
    void _connect();

    //* connection state changed
    void _connectionStateChanged( bool );

    //* connection lost
    void _lostConnection( const QString& );

    //* close job dialogs
    void _closeJobInformationDialogs();

    //* request default user
    void _requestDefaultUser();

    //* request user list
    void _requestUserList();

    //* request job list
    void _requestJobList();

    //* store default user
    void _updateDefaultUser( QString );

    //* process user list
    void _processUserList( UserSet );

    //* process job list
    void _processMasterList( QString, const JobSet& );

    //* process job list
    void _processChildrenList( Job, const JobSet& );

     //* process completed command
    void _commandCompleted( const Alien::Command& );

    //* commands failed
    void _commandsFailed( const Alien::Command::List& );

    //* busy
    void _busy()
    { busyWidget_->start(); }

    //* busy
    void _idle()
    { busyWidget_->stop(); }

    //* update actions visibility
    void _updateActions();

    //* update clear action
    void _updateClearAction();

    //* update user
    void _updateUser( QString );

    //* update status bar
    void _updateStatusBar( QString );

    //* active frame changed
    void _activeFrameChanged( int );

    //* toggle pause
    void _togglePause( bool );

    //* toggle summary
    void _toggleSummary( bool );

    //* add frame
    void _addFrame();

    //* remove frame
    void _removeFrame();

    //* remove frame at index
    void _removeFrame( int );

    //* request children list
    void _requestChildrenList( const Job& );

    //* retrace jobs
    void _requestJobInformation();

    //* kill jobs
    void _kill();

    //* resubmit jobs
    void _resubmit();

    //* clear completed jobs
    void _clear();

    //* show command list
    void _showCommandList();

    //* show log
    void _showLog();

    //* toggle log
    void _toggleLog( bool );

    //* toggle hide killed jobs
    void _toggleHideKilledJobs( bool );

    //* log dialog hidden
    void _logDialogHidden();

    //* command list destroyed
    void _commandListDestroyed();

    //* update configuration
    void _updateConfiguration();

    //* save configuration
    void _saveConfiguration();

    private:

    //* active frame
    bool _hasActiveFrame() const
    { return (bool) activeFrame_; }

    //* active frame
    JobManagerFrame& _activeFrame() const
    { return *activeFrame_; }

    //* true if frame is active
    bool _isActiveFrame( const JobManagerFrame& frame ) const
    { return &frame == activeFrame_; }

    //* find frame matching user
    /** return 0 if none found */
    JobManagerFrame* _findFrame( const QString& user ) const;

    //* active frame
    bool _activeFrameIsDefault() const;

    //* request job list
    void _requestJobList( QString );

    //* request job list
    void _requestJobList( QStringList );

    //* request clear queue
    void _requestClearQueue( QString );

    //* active frame
    void _setActiveFrame( JobManagerFrame& frame );

    //* create a new frame
    JobManagerFrame& _newFrame();

    //* get list of all active users
    UserSet _users() const;

    //* true if default user is set
    bool _hasDefaultUser() const
    { return !defaultUser_.isEmpty(); }

    //* true if a master job can be resubmitted
    bool _canResubmit( const Job& ) const;

    //* get commands associated to master job resubmission
    Alien::Command::List _masterJobResubmissionCommands( const Job& ) const;

    //* summary
    void _installActions();

    //* pointer to top interface
    Alien::Interface* interface_ = nullptr;

    //* command model
    Alien::CommandModel model_;

    //* log dialog
    LogDialog* logDialog_ = nullptr;

    //* toolbar
    CustomToolBar* userToolBar_ = nullptr;

    //* toolbar
    CustomToolBar* toolbar_ = nullptr;

    //* user
    UserSelectionFrame* userSelectionFrame_ = nullptr;

    //* status bar
    ProgressStatusBar* statusBar_ = nullptr;

    //* tab widget
    QTabWidget* tabWidget_ = nullptr;

    //* active frame
    JobManagerFrame* activeFrame_ = nullptr;

    //* busy widget
    BusyWidget* busyWidget_ = nullptr;

    //* refresh rate
    int refreshRate_ = 0;

    //* timer object for delayed tab change update
    QBasicTimer delayTimer_;

    //* timer object for scheduled update
    QBasicTimer timer_;

    //* default user
    QString defaultUser_;

    //* true if auto resubmission is requested
    bool autoResubmissionEnabled_ = false;

    //* maximum resubmition
    int maxResubmissions_ = 0;

    //*@name actions
    //@{

    //* connection
    QAction* connectAction_ = nullptr;

    //* command list
    QAction* commandListAction_ = nullptr;

    //* log action
    QAction* logAction_ = nullptr;

    //* log action
    QAction* toggleLogAction_ = nullptr;

    //* hide killed jobs
    QAction* hideKilledJobsAction_ = nullptr;

    //* pause
    QAction* pauseAction_ = nullptr;

    //* default user action
    QAction* defaultUserAction_ = nullptr;

    //* user list action
    QAction* userListAction_ = nullptr;

    //* update
    QAction* updateAction_ = nullptr;

    //* retrace
    QAction* traceAction_ = nullptr;

    //* kill action
    QAction* killAction_ = nullptr;

    //* resubmit
    QAction* resubmitAction_ = nullptr;

    //* toggle summary frame
    QAction* toggleSummaryAction_ = nullptr;

    //* add frame
    QAction* addFrameAction_ = nullptr;

    //* remove frame
    QAction* removeFrameAction_ = nullptr;

    //* reset time stamps
    QAction* resetTimeStampsAction_ = nullptr;

    //* clear
    QAction* clearAction_ = nullptr;

    //* abort
    QAction* abortAction_ = nullptr;

    //* close job dialogs
    QAction* closeJobDialogsAction_ = nullptr;

    //@}

    //* true if connection was valid at least once
    bool wasConnected_ = false;

};

#endif
