/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "OpenFileDialog.h"

#include "BaseIconNames.h"
#include "Debug.h"
#include "FileDialog.h"
#include "IconEngine.h"
#include "OpenWithComboBox.h"
#include "XmlOptions.h"

#include <QGridLayout>
#include <QButtonGroup>
#include <QToolButton>
#include <QLabel>

//___________________________________________________________________________
OpenFileDialog::OpenFileDialog( QWidget* parent, File file, Commands mode ):
    CustomDialog( parent, OkButton|CancelButton|Separator )
{

    Debug::Throw( "OpenFileDialog::OpenFileDialog.\n" );
    setOptionName( "OPEN_FILE_DIALOG" );

    // try load Question icon
    QHBoxLayout *hLayout( new QHBoxLayout );
    hLayout->setSpacing(10);
    hLayout->setMargin(0);
    mainLayout().addLayout( hLayout );
    {
        // icon
        QLabel* label = new QLabel( this );
        QIcon icon = IconEngine::get( IconNames::DialogWarning );
        label->setPixmap( icon.pixmap( iconSize() ) );
        hLayout->addWidget( label, 0 );

    }

    QVBoxLayout* vLayout = new QVBoxLayout;
    vLayout->setMargin(0);
    hLayout->addLayout( vLayout, 1 );

    vLayout->addStretch();
    vLayout->addWidget( new QLabel( tr( "Open file \"%1\"" ).arg( file.localName() ), this ) );
    vLayout->addWidget( new QLabel( "Which action would you like to perform ?", this ) );
    vLayout->addStretch();

    QGridLayout* gridLayout = new QGridLayout;
    gridLayout->setMargin(0);
    mainLayout().addLayout( gridLayout );

    gridLayout->addWidget( openWithRadioButton_ = new QRadioButton( "Open with", this ), 0, 0, 1, 1 );

    {
        gridLayout->addWidget( comboBox_ = new OpenWithComboBox( this ), 0, 1, 1, 1 );

        // retrieve applications from options
        const QString optionName = "APPLICATIONS";
        const auto applications( XmlOptions::get().specialOptions( optionName ) );
        for( const auto& option:applications )
        { comboBox_->addItem( File( option.raw() ) ); }

        // sytem default
        comboBox_->insertItem(0, tr( "System Default" ) );
        comboBox_->setCurrentIndex( 0 );
    }

    gridLayout->addWidget( saveAsRadioButton_ = new QRadioButton( "Save File", this ), 1, 0, 1, 3 );
    gridLayout->setColumnStretch( 1, 1 );

    QButtonGroup* buttonGroup = new QButtonGroup( this );
    buttonGroup->addButton( openWithRadioButton_ );
    buttonGroup->addButton( saveAsRadioButton_ );

    connect( openWithRadioButton_, SIGNAL(toggled(bool)), comboBox_, SLOT(setEnabled(bool)) );

    openWithRadioButton_->setEnabled( mode & OpenWith );
    saveAsRadioButton_->setEnabled( mode & SaveAs );

    if( mode & SaveAs )
    {

        saveAsRadioButton_->setChecked( true );
        comboBox_->setEnabled( false );

    } else {

        openWithRadioButton_->setChecked( true );

    }

    // connection
    connect( &okButton(), SIGNAL(clicked()), SLOT(_saveCommands()) );

}

//____________________________________________________________________________
bool OpenFileDialog::isOpenWithCommandValid() const
{ return comboBox_->isItemValid(); }

//____________________________________________________________________________
bool OpenFileDialog::isOpenWithCommandDefault() const
{ return comboBox_->currentIndex() == 0; }

//______________________________________________________
QString OpenFileDialog::openWithCommand() const
{ return comboBox_->command(); }

//____________________________________________________________________________
void OpenFileDialog::_saveCommands()
{

    const QString optionName = "APPLICATIONS";
    for( const auto& command:comboBox_->newItems() )
    { XmlOptions::get().add( optionName, Option( command, Option::Flag::Recordable|Option::Flag::Current ) ); }

}
