/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "SummaryBar.h"

#include "Color.h"
#include "Debug.h"
#include "Singleton.h"
#include "XmlOptions.h"

#include <QStyleOptionProgressBarV2>
#include <QStylePainter>

//___________________________________________________________
SummaryBar::SummaryBar( QWidget* parent ):
    QProgressBar( parent ),
    Counter( "SummaryBar" )
{
    Debug::Throw( "SummaryBar::SummaryBar.\n" );
    setSizePolicy( QSizePolicy::Minimum, QSizePolicy::Fixed );
    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    setTextVisible( false );
    _updateConfiguration();
}

//___________________________________________________________
void SummaryBar::updateJobStatistics( JobStatistics statistics )
{
    Debug::Throw( "SummaryBar::updateJobStatistics" );
    statistics_ = statistics;
    update();
}

//___________________________________________________________
void SummaryBar::paintEvent( QPaintEvent* event )
{
    QStylePainter painter(this);
    painter.setClipRegion( event->region() );
    QStyleOptionProgressBarV2 opt;
    initStyleOption(&opt);
    opt.text = "";
    opt.minimum = 0;
    opt.maximum = qMax<int>(statistics_.get( JobStatistics::Total ), 1);

    // draw container
    QStyleOptionProgressBarV2 subOpt(opt);
    subOpt.rect = style()->subElementRect( QStyle::SE_ProgressBarGroove, &opt, this);
    style()->drawControl( QStyle::CE_ProgressBarGroove, &subOpt, &painter, this);
    if( !statistics_.get( JobStatistics::Total ) ) return;

    #if defined(Q_OS_MAC)

    // for mac style, the progressbar color cannot be changed (because native macintosh methods are called).
    // as a fall-back solution, the progressbar only represents the number of completed jobs
    subOpt.progress = statistics_.get( JobStatistics::Completed );
    subOpt.rect = opt.rect;
    subOpt.rect = style()->subElementRect( QStyle::SE_ProgressBarContents, &subOpt, this);
    style()->drawControl( QStyle::CE_ProgressBarContents, &subOpt, &painter, this);

    #else

    // draw running jobs
    if( statistics_.get( JobStatistics::Running ) > 0 )
    {
        subOpt.progress = statistics_.get( JobStatistics::Running ) + statistics_.get( JobStatistics::Failed ) + statistics_.get( JobStatistics::Completed );
        subOpt.rect = opt.rect;        subOpt.rect = style()->subElementRect( QStyle::SE_ProgressBarContents, &subOpt, this);
        subOpt.palette.setColor( QPalette::Highlight, runningJobColor_ );
        style()->drawControl( QStyle::CE_ProgressBarContents, &subOpt, &painter, this);
    }

    // draw failed jobs
    if( statistics_.get( JobStatistics::Failed ) > 0 )
    {
        subOpt.progress = statistics_.get( JobStatistics::Completed ) + statistics_.get( JobStatistics::Failed );
        subOpt.rect = opt.rect;
        subOpt.rect = style()->subElementRect( QStyle::SE_ProgressBarContents, &subOpt, this);
        subOpt.palette.setColor( QPalette::Highlight, failedJobColor_ );
        style()->drawControl( QStyle::CE_ProgressBarContents, &subOpt, &painter, this);
    }

    // draw completed jobs
    if( statistics_.get( JobStatistics::Completed ) > 0 )
    {
        const int value( statistics_.get( JobStatistics::Completed ) );
        subOpt.progress = value;
        subOpt.rect = opt.rect;        subOpt.rect = style()->subElementRect( QStyle::SE_ProgressBarContents, &subOpt, this);
        subOpt.palette.setColor( QPalette::Highlight, completedJobColor_ );
        style()->drawControl( QStyle::CE_ProgressBarContents, &subOpt, &painter, this);
    }
    #endif

    painter.end();
    return;

}

//___________________________________________________________
void SummaryBar::_updateConfiguration()
{
    Debug::Throw( "SummaryBar::_updateConfiguration.\n" );

    // running job color
    runningJobColor_ = Base::Color( XmlOptions::get().get<Base::Color>("RUNNING_JOB_COLOR") );
    if( !runningJobColor_.isValid() ) runningJobColor_ = palette().color( QPalette::Highlight );

    // completed job color
    completedJobColor_ = Base::Color( XmlOptions::get().get<Base::Color>("COMPLETED_JOB_COLOR") );
    if( !completedJobColor_.isValid() ) completedJobColor_ = palette().color( QPalette::Highlight );

    // failed job color
    failedJobColor_ = Base::Color( XmlOptions::get().get<Base::Color>("FAILED_JOB_COLOR") );
    if( !failedJobColor_.isValid() ) failedJobColor_ = palette().color( QPalette::Highlight );

}
